r"""Implement handlers for ``polars.DataFrame``s, ``polars.LazyFrame``s,
and ``polars.Series``es."""

from __future__ import annotations

__all__ = ["PolarsDataFrameEqualHandler", "PolarsLazyFrameEqualHandler", "PolarsSeriesEqualHandler"]

import inspect
import logging
from functools import lru_cache
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.utils.imports import is_polars_available

if TYPE_CHECKING or is_polars_available():
    import polars as pl
    import polars.selectors as cs
    from polars import testing
else:  # pragma: no cover
    from coola.utils.fallback.polars import polars as pl

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class PolarsDataFrameEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two ``polars.DataFrame``s are equal.

    This handler returns ``True`` if the two ``polars.DataFrame``s are
    equal, otherwise ``False``. This handler is designed to be used
    at the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import polars as pl
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PolarsDataFrameEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PolarsDataFrameEqualHandler()
        >>> handler.handle(
        ...     pl.DataFrame({"col": [1, 2, 3]}),
        ...     pl.DataFrame({"col": [1, 2, 3]}),
        ...     config,
        ... )
        True
        >>> handler.handle(
        ...     pl.DataFrame({"col": [1, 2, 3]}),
        ...     pl.DataFrame({"col": [1, 2, 4]}),
        ...     config,
        ... )
        False

        ```
    """

    def handle(
        self,
        actual: pl.DataFrame,
        expected: pl.DataFrame,
        config: EqualityConfig,
    ) -> bool:
        object_equal = frame_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="polars.DataFrames")
            )
        return object_equal


class PolarsLazyFrameEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two ``polars.LazyFrame``s are equal.

    This handler returns ``True`` if the two ``polars.LazyFrame``s are
    equal, otherwise ``False``. This handler is designed to be used
    at the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import polars as pl
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PolarsLazyFrameEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PolarsLazyFrameEqualHandler()
        >>> handler.handle(
        ...     pl.LazyFrame({"col": [1, 2, 3]}),
        ...     pl.LazyFrame({"col": [1, 2, 3]}),
        ...     config,
        ... )
        True
        >>> handler.handle(
        ...     pl.LazyFrame({"col": [1, 2, 3]}),
        ...     pl.LazyFrame({"col": [1, 2, 4]}),
        ...     config,
        ... )
        False

        ```
    """

    def handle(
        self,
        actual: pl.LazyFrame,
        expected: pl.LazyFrame,
        config: EqualityConfig,
    ) -> bool:
        object_equal = frame_equal(actual.collect(), expected.collect(), config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="polars.LazyFrames")
            )
        return object_equal


class PolarsSeriesEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two ``polars.Series``es are equal.

    This handler returns ``True`` if the two ``polars.Series``es are
    equal, otherwise ``False``. This handler is designed to be used
    at the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import polars as pl
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PolarsSeriesEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PolarsSeriesEqualHandler()
        >>> handler.handle(pl.Series([1, 2, 3]), pl.Series([1, 2, 3]), config)
        True
        >>> handler.handle(pl.Series([1, 2, 3]), pl.Series([1, 2, 4]), config)
        False

        ```
    """

    def handle(
        self,
        actual: pl.Series,
        expected: pl.Series,
        config: EqualityConfig,
    ) -> bool:
        object_equal = series_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="polars.Series")
            )
        return object_equal


def has_nan(df_or_series: pl.DataFrame | pl.Series) -> bool:
    r"""Indicate if a DataFrame or Series has NaN values.

    Args:
        df_or_series: The DataFrame or series to check.

    Returns:
        ``True`` if the DataFrame or Series has NaN values,
            otherwise ``False``.
    """
    if isinstance(df_or_series, pl.Series):
        return df_or_series.dtype.is_numeric() and df_or_series.is_nan().any()
    frame = df_or_series.select(cs.numeric())
    if frame.is_empty():  # This if is necessary for previous polars version (1.0)
        return False
    return frame.select(pl.any_horizontal(pl.all().is_nan().any())).item()


def frame_equal(df1: pl.DataFrame, df2: pl.DataFrame, config: EqualityConfig) -> bool:
    r"""Indicate if the two DataFrames are equal or not.

    Args:
        df1: The first DataFrame to compare.
        df2: The second DataFrame to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two DataFrames are equal, otherwise ``False``.
    """
    if not config.equal_nan and has_nan(df1):
        return False
    try:
        assert_frame_equal(df1, df2, atol=config.atol, rtol=config.rtol)
    except AssertionError:
        return False
    return True


def series_equal(series1: pl.Series, series2: pl.Series, config: EqualityConfig) -> bool:
    r"""Indicate if the two series are equal or not.

    Args:
        series1: The first series to compare.
        series2: The second series to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two series are equal, otherwise ``False``.
    """
    if not config.equal_nan and has_nan(series1):
        return False
    try:
        assert_series_equal(
            series1,
            series2,
            atol=config.atol,
            rtol=config.rtol,
        )
    except AssertionError:
        return False
    return True


def assert_frame_equal(
    df1: pl.DataFrame,
    df2: pl.DataFrame,
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> None:
    r"""Compare two polars DataFrames with tolerance.

    This function supports multiple polars versions.

    Args:
        df1: The first DataFrame to compare.
        df2: The second DataFrame to compare.
        rtol: The relative tolerance parameter.
        atol: The absolute tolerance parameter.
    """
    check_exact = atol == 0 and rtol == 0
    if is_new_naming():
        testing.assert_frame_equal(df1, df2, abs_tol=atol, rel_tol=rtol, check_exact=check_exact)
    else:
        testing.assert_frame_equal(df1, df2, atol=atol, rtol=rtol, check_exact=check_exact)


def assert_series_equal(
    series1: pl.Series,
    series2: pl.Series,
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> None:
    r"""Compare two polars Series with tolerance.

    This function supports multiple polars versions.

    Args:
        series1: The first series to compare.
        series2: The second series to compare.
        rtol: The relative tolerance parameter.
        atol: The absolute tolerance parameter.
    """
    check_exact = atol == 0 and rtol == 0
    if is_new_naming():
        testing.assert_series_equal(
            series1, series2, abs_tol=atol, rel_tol=rtol, check_exact=check_exact
        )
    else:
        testing.assert_series_equal(series1, series2, atol=atol, rtol=rtol, check_exact=check_exact)


@lru_cache
def is_new_naming() -> bool:
    r"""Indicate if polars uses the new naming for ``abs_tol`` and
    ``rel_tol``.

    Returns:
        ``True`` if ``abs_tol`` and ``rel_tol`` are used,
            otherwise ``atol`` and ``rtol`` are used.
    """
    return "abs_tol" in inspect.signature(pl.testing.assert_frame_equal).parameters



r"""Contain the reducer implementations."""

__all__ = [
    "BaseBasicReducer",
    "BaseReducer",
    "EmptySequenceError",
    "NativeReducer",
    "NumpyReducer",
    "TorchReducer",
    "auto_reducer",
]

from coola.reducer.base import BaseBasicReducer, BaseReducer, EmptySequenceError
from coola.reducer.native import NativeReducer
from coola.reducer.numpy import NumpyReducer
from coola.reducer.torch import TorchReducer
from coola.reducer.utils import auto_reducer



r"""Define the abstract base class for child finders used in breadth-
first search."""

from __future__ import annotations

__all__ = ["BaseChildFinder"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

T = TypeVar("T")


class BaseChildFinder(ABC, Generic[T]):
    r"""Abstract base class for child finders used in breadth-first
    search.

    This class defines the interface that all child finders must implement.
    Child finders are responsible for finding and yielding the immediate children
    of a given data structure. Custom child finders can be registered with a
    ``ChildFinderRegistry`` to handle specific data types during BFS traversal.

    The generic type parameter ``T`` indicates the type of data this child finder
    is designed to handle.

    Notes:
        - Subclasses must implement the ``find_children`` method.
        - For leaf types (types with no children), simply return without yielding.

    Example:
        ```pycon
        >>> from coola.iterator.bfs import DefaultChildFinder
        >>> child_finder = DefaultChildFinder()
        >>> list(child_finder.find_children(42))
        []
        >>> list(child_finder.find_children("hello"))
        []

        ```
    """

    @abstractmethod
    def find_children(self, data: T) -> Iterator[Any]:
        r"""Find and yield the immediate children of the given data
        structure.

        This method defines how to extract children from the data structure.
        For container types, this typically means yielding the contained elements.
        For leaf types, this method should return without yielding anything.

        Args:
            data: The data structure whose children should be found.

        Yields:
            The immediate children of the data structure. The type of yielded
            elements depends on the specific data structure being processed.

        Notes:
            - This method should only yield direct children, not recurse deeply.
            - The BFS traversal logic handles visiting children recursively.

        Example:
            ```pycon
            >>> from coola.iterator.bfs import DefaultChildFinder
            >>> child_finder = DefaultChildFinder()
            >>> list(child_finder.find_children(42))
            []
            >>> list(child_finder.find_children("hello"))
            []

            ```
        """



r"""Implement some utility functions to manage environment variables."""

from __future__ import annotations

__all__ = ["temp_env_vars"]

import os
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Generator


@contextmanager
def temp_env_vars(env_vars: dict[str, Any]) -> Generator[None, None, None]:
    r"""Context manager to temporarily set or modify environment
    variables.

    This context manager allows you to temporarily change environment variables
    within a specific scope. All changes are automatically reverted when exiting
    the context, even if an exception occurs.

    Args:
        env_vars: Environment variables to set as keyword arguments.
            Keys are variable names, values are the values to set.
            Values are automatically converted to strings.

    Behavior:
        - If a variable already exists, its original value is saved and restored
        - If a variable doesn't exist, it's created temporarily and removed on exit
        - All operations are guaranteed to execute via try/finally
        - Thread-safe for the current process (but note that os.environ affects
          the entire process, not just the current thread)

    Example:
        ```pycon
        >>> from coola.utils.env_vars import temp_env_vars
        >>> # Temporarily override an existing variable
        >>> os.environ["HOME"] = "/original/home"
        >>> with temp_env_vars({"HOME": "/tmp/home"}):
        ...     print(os.environ["HOME"])  # '/tmp/home'
        ...
        >>> print(os.environ["HOME"])  # '/original/home'
        >>> # Temporarily create new variables
        >>> with temp_env_vars({"API_KEY": "secret123", "DEBUG": "true"}):
        ...     print(os.environ["API_KEY"])  # 'secret123'
        ...     print(os.environ["DEBUG"])  # 'true'
        ...
        >>> print(os.environ.get("API_KEY"))  # None (removed)

        ```

    Notes:
        Changes to os.environ affect the entire Python process, not just the
        current thread. Use with caution in multi-threaded applications.
    """
    # Store original values (or None if they didn't exist)
    original = {key: os.environ.get(key, None) for key in env_vars}

    # Set new values
    for key, value in env_vars.items():
        os.environ[key] = str(value)

    try:
        yield
    finally:
        # Restore original state
        for key, value in original.items():
            if value is None:
                # Remove if it didn't exist before
                os.environ.pop(key)
            else:
                # Restore original value
                os.environ[key] = value



r"""Implement the NDArray summarizer for NumPy arrays.

This module provides a summarizer that creates compact, informative
string representations of NumPy ndarrays, showing metadata like shape
and dtype instead of the full data by default.
"""

from __future__ import annotations

__all__ = ["NDArraySummarizer"]

from typing import TYPE_CHECKING

from coola.summary.base import BaseSummarizer
from coola.utils.imports import check_numpy, is_numpy_available

if TYPE_CHECKING:
    from coola.summary.registry import SummarizerRegistry

if TYPE_CHECKING or is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    from coola.utils.fallback.numpy import numpy as np


class NDArraySummarizer(BaseSummarizer[np.ndarray]):
    r"""Implement a summarizer for ``numpy.ndarray`` objects.

    This summarizer generates compact string representations of NumPy
    arrays. By default, it displays metadata (type, shape, dtype)
    rather than array values, making it suitable for logging and debugging
    large arrays. Optionally, it can show the full array representation.

    Args:
        show_data: If ``True``, returns the default array string
            representation (same as ``repr(array)``), displaying
            actual values. If ``False`` (default), returns only
            metadata in a compact format:
            ``<class> | shape=<shape> | dtype=<dtype>``.
            Default: ``False``

    Raises:
        RuntimeError: If NumPy is not installed or available.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.summary import SummarizerRegistry, NDArraySummarizer
        >>> registry = SummarizerRegistry()

        >>> # Default behavior: show metadata only
        >>> summarizer = NDArraySummarizer()
        >>> print(summarizer.summarize(np.arange(11), registry))
        <class 'numpy.ndarray'> | shape=(11,) | dtype=int64

        >>> # Works with arrays of any shape and dtype
        >>> print(summarizer.summarize(np.ones((2, 3, 4)), registry))
        <class 'numpy.ndarray'> | shape=(2, 3, 4) | dtype=float64

        >>> # Show full array data
        >>> summarizer = NDArraySummarizer(show_data=True)
        >>> print(summarizer.summarize(np.arange(11), registry))
        array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10])

        ```
    """

    def __init__(self, show_data: bool = False) -> None:
        check_numpy()
        self._show_data = bool(show_data)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(show_data={self._show_data})"

    def equal(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        return self._show_data == other._show_data

    def summarize(
        self,
        data: np.ndarray,
        registry: SummarizerRegistry,  # noqa: ARG002
        depth: int = 0,  # noqa: ARG002
        max_depth: int = 1,  # noqa: ARG002
    ) -> str:
        if self._show_data:
            return repr(data)
        return " | ".join([f"{type(data)}", f"shape={data.shape}", f"dtype={data.dtype}"])



r"""Implement handlers for sequence objects."""

from __future__ import annotations

__all__ = ["SequenceSameValuesHandler"]

import logging
from typing import TYPE_CHECKING, Any

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_sequence_difference
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.equality.handler.utils import check_recursion_depth

if TYPE_CHECKING:
    from collections.abc import Sequence

    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class SequenceSameValuesHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two sequences have the same values.

    This handler returns ``False`` if the two sequences have at least
    one different value, otherwise it passes the inputs to the next
    handler. If the sequences have different length, this handler
    checks only the values of the shortest sequence.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SequenceSameValuesHandler, TrueHandler
        >>> config = EqualityConfig()
        >>> handler = SequenceSameValuesHandler(next_handler=TrueHandler())
        >>> handler.handle([1, 2, 3], [1, 2, 3], config)
        True
        >>> handler.handle([1, 2, 3], [1, 2, 4], config)
        False

        ```
    """

    def handle(
        self,
        actual: Sequence[Any],
        expected: Sequence[Any],
        config: EqualityConfig,
    ) -> bool:
        with check_recursion_depth(config):
            for idx, (value1, value2) in enumerate(zip(actual, expected)):
                if not config.registry.objects_are_equal(value1, value2, config):
                    if config.show_difference:
                        logger.info(
                            format_sequence_difference(
                                actual,
                                expected,
                                different_index=idx,
                            )
                        )
                    return False
            return self._handle_next(actual, expected, config=config)



r"""Define the default child finder class for BFS traversal."""

from __future__ import annotations

__all__ = ["DefaultChildFinder"]

from typing import TYPE_CHECKING, Any

from coola.iterator.bfs.base import BaseChildFinder

if TYPE_CHECKING:
    from collections.abc import Iterator


class DefaultChildFinder(BaseChildFinder[Any]):
    r"""Default child finder for breadth-first search traversal.

        This child finder serves as the fallback handler for objects that don't have
        a specialized child finder registered. It treats the input data as a leaf node
        with no children, so it yields nothing during traversal.

        The DefaultChildFinder is typically used for:
        - Primitive types (int, float, str, bool, None)
        - Objects without internal structure to traverse
        - Terminal nodes in a data structure

    Example:
        ```pycon
        >>> from coola.iterator.bfs import DefaultChildFinder
        >>> child_finder = DefaultChildFinder()
        >>> list(child_finder.find_children(42))
        []
        >>> list(child_finder.find_children("hello"))
        []

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def find_children(self, data: object) -> Iterator[Any]:  # noqa: ARG002
        return
        yield



r"""Define the public interface to recursively iterate over nested
data."""

from __future__ import annotations

__all__ = ["dfs_iterate", "get_default_registry", "register_iterators"]

from collections.abc import Iterable, Iterator, Mapping
from typing import TYPE_CHECKING, Any

from coola.iterator.dfs.default import DefaultIterator
from coola.iterator.dfs.iterable import IterableIterator
from coola.iterator.dfs.mapping import MappingIterator
from coola.iterator.dfs.registry import IteratorRegistry

if TYPE_CHECKING:
    from coola.iterator.dfs.base import BaseIterator


def dfs_iterate(data: object, registry: IteratorRegistry | None = None) -> Iterator[Any]:
    """Perform Depth-First Search (DFS) iteration over nested data
    structures (lists, dicts, tuples, sets, etc.).

    This function yields elements from the data structure in a DFS manner, recursively
    traversing all levels of nested structures. It uses the appropriate iterators registered
    for the data types (e.g., lists, dictionaries, etc.).

    Args:
        data: The nested data structure to traverse. Can be a list, tuple, dict, set, or any
            other registered type.
        registry: The registry to resolve iterators for nested data. If `None`, the default
            registry is used.

    Yields:
        The elements from the nested data structure in DFS order.

    Example:
        ```pycon
        >>> from coola.iterator import dfs_iterate
        >>> list(dfs_iterate({"a": 1, "b": "abc"}))
        [1, 'abc']
        >>> list(dfs_iterate([1, [2, 3], {"x": 4}]))
        [1, 2, 3, 4]

        ```
    """
    if registry is None:
        registry = get_default_registry()
    yield from registry.iterate(data)


def register_iterators(
    mapping: Mapping[type, BaseIterator[Any]],
    exist_ok: bool = False,
) -> None:
    """Register custom iterators to the default global registry.

    This allows users to add support for custom types without modifying global state directly.

    Args:
        mapping: A dictionary mapping Python types to their corresponding iterator instances.
        exist_ok: If `True`, existing registrations for types will be overwritten.
            If `False`, an error is raised when a type is already registered.

    Example:
        ```pycon
        >>> from coola.iterator.dfs import register_iterators, IterableIterator, IteratorRegistry
        >>> register_iterators({list: IterableIterator()}, exist_ok=True)
        >>> registry = get_default_registry()
        >>> list(registry.iterate([1, 2, 3]))
        [1, 2, 3]

        ```
    """
    get_default_registry().register_many(mapping, exist_ok=exist_ok)


def get_default_registry() -> IteratorRegistry:
    """Get or create the default global registry for iterators.

    This function returns a singleton instance of the `IteratorRegistry`, which is
    pre-configured with iterators for common Python types, including iterables (lists,
    tuples), mappings (dicts), sets, and scalars (int, float, str, bool). The registry
    is used to look up the appropriate iterator for a given data structure during iteration.

    Returns:
        An `IteratorRegistry` instance with iterators registered for common Python types.

    Notes:
        The singleton pattern means any changes to the returned registry affect all future
        calls to this function. If an isolated registry is needed, create a new `IteratorRegistry`
        instance directly.

    Example:
        ```pycon
        >>> from coola.iterator.dfs import get_default_registry
        >>> reg = get_default_registry()
        >>> list(reg.iterate([1, 2, 3]))
        [1, 2, 3]

        ```
    """
    if not hasattr(get_default_registry, "_registry"):
        registry = IteratorRegistry()
        _register_default_iterators(registry)
        get_default_registry._registry = registry
    return get_default_registry._registry


def _register_default_iterators(registry: IteratorRegistry) -> None:
    """Register default iterators for common Python types.

    This internal function registers the standard type-to-iterator mappings that are used
    by the default registry. The registration ensures that each type is handled appropriately
    during iteration, including handling nested structures.

    Args:
        registry: The `IteratorRegistry` to populate with default iterators.

    Notes:
        This function is automatically called by `get_default_registry()` and should not
        be called directly by users.
    """
    default_iterator = DefaultIterator()
    iterable_iterator = IterableIterator()
    mapping_iterator = MappingIterator()

    registry.register_many(
        {
            # Scalar types - no recursion needed
            object: default_iterator,
            str: default_iterator,  # Strings should not be iterated character by character
            bytes: default_iterator,
            int: default_iterator,
            float: default_iterator,
            complex: default_iterator,
            bool: default_iterator,
            # Iterables - recursive iteration (lists, tuples, etc.)
            list: iterable_iterator,
            tuple: iterable_iterator,
            range: iterable_iterator,
            set: iterable_iterator,
            frozenset: iterable_iterator,
            Iterable: iterable_iterator,
            # Mappings - recursive iteration (dictionaries)
            dict: mapping_iterator,
            Mapping: mapping_iterator,
        }
    )



r"""Define a thread-safe registry."""

from __future__ import annotations

__all__ = ["Registry"]

import threading
from typing import TYPE_CHECKING, Generic, TypeVar

from coola.equality.interface import objects_are_equal
from coola.utils.format import repr_indent, repr_mapping, str_indent, str_mapping

if TYPE_CHECKING:
    from collections.abc import ItemsView, Iterator, KeysView, Mapping, ValuesView

K = TypeVar("K")  # Key type
V = TypeVar("V")  # Value type


class Registry(Generic[K, V]):
    r"""A thread-safe generic key-value registry for storing and managing
    typed mappings.

    The Registry class provides a type-safe container for registering and
    retrieving values by key. It supports all standard dictionary operations
    through operator overloading and provides additional methods for safe
    registration and querying. All operations are protected by a lock to
    ensure thread safety in concurrent environments.

    Args:
        initial_state: An optional dictionary to initialize the registry with.
            If provided, a copy is made to prevent external modifications.
            Defaults to None, which creates an empty registry.

    Attributes:
        _state: Internal dictionary storing the key-value pairs.
        _lock: Threading lock for synchronizing access to the registry.

    Example:
        Basic usage with registration and retrieval:

        ```pycon
        >>> from coola.registry import Registry
        >>> registry = Registry[str, int]()
        >>> registry.register("key1", 42)
        >>> registry.get("key1")
        42
        >>> registry
        Registry(
          (key1): 42
        )

        ```

        Using dictionary-style operations:

        ```pycon
        >>> from coola.registry import Registry
        >>> registry = Registry[str, int]()
        >>> registry["key2"] = 100
        >>> "key2" in registry
        True
        >>> del registry["key2"]

        ```

        Initializing with existing data:

        ```pycon
        >>> from coola.registry import Registry
        >>> registry = Registry[str, int](initial_state={"a": 1, "b": 2})
        >>> len(registry)
        2

        ```
    """

    def __init__(self, initial_state: dict[K, V] | None = None) -> None:
        self._state: dict[K, V] = initial_state.copy() if initial_state else {}
        self._lock: threading.RLock = threading.RLock()  # RLock allows re-entrant locking

    def __contains__(self, key: K) -> bool:
        return self.has(key)

    def __getitem__(self, key: K) -> V:
        with self._lock:
            if key not in self._state:
                msg = f"Key '{key}' is not registered"
                raise KeyError(msg)
            return self._state[key]

    def __setitem__(self, key: K, value: V) -> None:
        self.register(key, value, exist_ok=True)

    def __delitem__(self, key: K) -> None:
        self.unregister(key)

    def __iter__(self) -> Iterator[K]:
        with self._lock:
            return iter(self._state.copy())

    def __len__(self) -> int:
        with self._lock:
            return len(self._state)

    def __repr__(self) -> str:
        with self._lock:
            snapshot = self._state.copy()
        return f"{self.__class__.__qualname__}(\n  {repr_indent(repr_mapping(snapshot))}\n)"

    def __str__(self) -> str:
        with self._lock:
            snapshot = self._state.copy()
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(snapshot))}\n)"

    def clear(self) -> None:
        """Remove all entries from the registry.

        This method empties the registry, leaving it in the same state as a
        newly created empty registry. This operation cannot be undone.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]({"key1": 42, "key2": 100})
            >>> len(registry)
            2
            >>> registry.clear()
            >>> len(registry)
            0
            >>> registry.has("key1")
            False

            ```
        """
        with self._lock:
            self._state.clear()

    def equal(self, other: object, equal_nan: bool = False) -> bool:
        r"""Indicate if two objects are equal or not.

        Args:
            other: The object to compare with.
            equal_nan: If ``True``, then two ``NaN``s will be
                considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry1 = Registry[str, int]({"key1": 42, "key2": 100})
            >>> registry2 = Registry[str, int]({"key1": 42, "key2": 100})
            >>> registry3 = Registry[str, int]({"key1": 42})
            >>> registry1.equal(registry2)
            True
            >>> registry1.equal(registry3)
            False

            ```
        """
        if type(other) is not type(self):
            return False

        # Acquire locks in a consistent order based on object id to avoid deadlock
        first, second = (self, other) if id(self) < id(other) else (other, self)
        with first._lock, second._lock:
            return objects_are_equal(self._state, other._state, equal_nan=equal_nan)

    def get(self, key: K, default: V | None = None) -> V | None:
        """Retrieve the value associated with a key.

        This method performs a lookup in the registry and returns the
        corresponding value.

        Args:
            key: The key whose value should be retrieved.
            default: Value to return if the key does not exist.

        Returns:
            The value associated with the specified key.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]({"key1": 42, "key2": 100})
            >>> registry.get("key1")
            42
            >>> registry.get("missing")
            None

            ```
        """
        with self._lock:
            return self._state.get(key, default)

    def has(self, key: K) -> bool:
        """Check whether a key is registered in the registry.

        This method provides a safe way to test for key existence without
        risking a KeyError exception.

        Args:
            key: The key to check for existence.

        Returns:
            True if the key exists in the registry, False otherwise.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]({"key1": 42, "key2": 100})
            >>> registry.has("key1")
            True
            >>> registry.has("missing")
            False

            ```
        """
        with self._lock:
            return key in self._state

    def register(self, key: K, value: V, exist_ok: bool = False) -> None:
        """Register a new key-value pair in the registry.

        By default, this method raises an error if you try to register a key
        that already exists. This prevents accidental overwriting of values.
        Set exist_ok=True to allow overwriting.

        Args:
            key: The key to register. Must be hashable.
            value: The value to associate with the key.
            exist_ok: Controls behavior when the key already exists.
                If False (default), raises RuntimeError for duplicate keys.
                If True, silently overwrites the existing value.

        Raises:
            RuntimeError: If the key is already registered and exist_ok is False.
                The error message provides guidance on how to resolve the conflict.

        Example:
            Basic registration:

            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]()
            >>> registry.register("key1", 42)
            >>> registry.get("key1")
            42

            ```

            Attempting to register a duplicate key:

            ```pycon
            >>> registry.register("key1", 100)  # doctest: +SKIP
            RuntimeError: A value is already registered for 'key1'...

            ```

            Overwriting with exist_ok:

            ```pycon
            >>> registry.register("key1", 100, exist_ok=True)
            >>> registry.get("key1")
            100

            ```
        """
        with self._lock:
            if key in self._state and not exist_ok:
                msg = (
                    f"A value is already registered for '{key}'. "
                    "Use a different key or set exist_ok=True to override."
                )
                raise RuntimeError(msg)
            self._state[key] = value

    def register_many(self, mapping: Mapping[K, V], exist_ok: bool = False) -> None:
        """Register multiple key-value pairs in a single operation.

        This is a convenience method for bulk registration. It iterates through
        the provided mapping and registers each key-value pair. All registrations
        follow the same exist_ok policy. The operation is atomic when exist_ok
        is False - if any key already exists, no changes are made.

        Args:
            mapping: A dictionary or mapping containing the key-value pairs
                to register. The keys and values must match the registry's
                type parameters.
            exist_ok: Controls behavior when any key already exists.
                If False (default), raises error on the first duplicate key.
                If True, overwrites all existing values without error.

        Raises:
            RuntimeError: If exist_ok is False and any key in the mapping
                is already registered. The error occurs on the first duplicate
                encountered, and no partial registration occurs.

        Example:
            Registering multiple entries at once:

            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]()
            >>> registry.register_many({"key1": 42, "key2": 100, "key3": 7})
            >>> registry.get("key1")
            42
            >>> registry.get("key2")
            100
            >>> len(registry)
            3

            ```

            Bulk update with exist_ok:

            ```pycon
            >>> registry.register_many({"key1": 1, "key4": 4}, exist_ok=True)
            >>> registry.get("key1")
            1
            >>> registry.get("key4")
            4

            ```
        """
        with self._lock:
            # Check all keys first if exist_ok is False
            if not exist_ok and (duplicates := set(mapping) & set(self._state)):
                msg = (
                    f"Keys already registered: {', '.join(map(str, duplicates))}. "
                    "Use different keys or set exist_ok=True to override."
                )
                raise RuntimeError(msg)
            self._state.update(mapping)

    def unregister(self, key: K) -> V:
        """Remove a key-value pair from the registry and return the
        value.

        This method removes the specified key from the registry and returns
        the value that was associated with it. This allows you to retrieve
        the value one last time before it's removed.

        Args:
            key: The key to unregister and remove from the registry.

        Returns:
            The value that was associated with the key before removal.

        Raises:
            KeyError: If the key is not registered. The error message
                includes the key that was not found.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int]({"key1": 42, "key2": 100})
            >>> registry.has("key1")
            True
            >>> value = registry.unregister("key1")
            >>> value
            42
            >>> registry.has("key1")
            False

            ```
        """
        with self._lock:
            if key not in self._state:
                msg = f"Key '{key}' is not registered"
                raise KeyError(msg)
            return self._state.pop(key)

    def items(self) -> ItemsView[K, V]:
        """Return key-value pairs.

        Returns:
            The key-value pairs.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int](initial_state={"a": 1, "b": 2})
            >>> registry.items()
            dict_items([('a', 1), ('b', 2)])

            ```
        """
        with self._lock:
            return self._state.copy().items()

    def keys(self) -> KeysView[K]:
        """Return registered keys.

        Returns:
            The registered keys.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int](initial_state={"a": 1, "b": 2})
            >>> registry.keys()
            dict_keys(['a', 'b'])

            ```
        """
        with self._lock:
            return self._state.copy().keys()

    def values(self) -> ValuesView[V]:
        """Return registered values.

        Returns:
            The registered values.

        Example:
            ```pycon
            >>> from coola.registry import Registry
            >>> registry = Registry[str, int](initial_state={"a": 1, "b": 2})
            >>> registry.values()
            dict_values([1, 2])

            ```
        """
        with self._lock:
            return self._state.copy().values()



r"""Contain utility functions for mappings."""

from __future__ import annotations

__all__ = ["sort_by_keys", "sort_by_values"]

import operator
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping


def sort_by_keys(mapping: Mapping[Any, Any]) -> dict[Any, Any]:
    r"""Sort a dictionary by keys.

    Args:
        mapping: The dictionary to sort.

    Returns:
        dict: The sorted dictionary.

    Example:
        ```pycon
        >>> from coola.utils.mapping import sort_by_keys
        >>> sort_by_keys({"dog": 1, "cat": 5, "fish": 2})
        {'cat': 5, 'dog': 1, 'fish': 2}

        ```
    """
    return dict(sorted(mapping.items()))


def sort_by_values(mapping: Mapping[Any, Any]) -> dict[Any, Any]:
    r"""Sort a dictionary by keys.

    Args:
        mapping: The dictionary to sort.

    Returns:
        dict: The sorted dictionary.

    Example:
        ```pycon
        >>> from coola.utils.mapping import sort_by_values
        >>> sort_by_values({"dog": 1, "cat": 5, "fish": 2})
        {'dog': 1, 'fish': 2, 'cat': 5}

        ```
    """
    return dict(sorted(mapping.items(), key=operator.itemgetter(1)))



r"""Contain fallback implementations used when ``polars`` dependency is
not available."""

from __future__ import annotations

__all__ = ["polars"]

from types import ModuleType
from typing import Any

from coola.utils.imports import raise_error_polars_missing


class FakeClass:
    r"""Fake class that raises an error because polars is not installed.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Raises:
        RuntimeError: polars is required for this functionality.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        raise_error_polars_missing()


# Create a fake polars package
polars: ModuleType = ModuleType("polars")
polars.DataFrame = FakeClass
polars.LazyFrame = FakeClass
polars.Series = FakeClass



r"""Define the BFS child finder class for handling iterable data."""

from __future__ import annotations

__all__ = ["IterableChildFinder"]

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from coola.iterator.bfs.base import BaseChildFinder

if TYPE_CHECKING:
    from collections.abc import Iterator


class IterableChildFinder(BaseChildFinder[Iterable[Any]]):
    r"""Child finder for iterable objects.

    This child finder handles iterable objects by yielding each element
    of the iterable. It works with lists, tuples, sets, strings, and any
    object implementing the Iterable protocol.

    Example:
        ```pycon
        >>> from coola.iterator.bfs import IterableChildFinder
        >>> child_finder = IterableChildFinder()
        >>> list(child_finder.find_children((4, 2, 1)))
        [4, 2, 1]
        >>> list(child_finder.find_children("hello"))
        ['h', 'e', 'l', 'l', 'o']

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def find_children(self, data: Iterable[Any]) -> Iterator[Any]:
        yield from data



r"""Contain fallback implementations used when ``pandas`` dependency is
not available."""

from __future__ import annotations

__all__ = ["pandas"]

from types import ModuleType
from typing import Any

from coola.utils.imports import raise_error_pandas_missing


class FakeClass:
    r"""Fake class that raises an error because pandas is not installed.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Raises:
        RuntimeError: pandas is required for this functionality.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        raise_error_pandas_missing()


# Create a fake pandas package
pandas: ModuleType = ModuleType("pandas")
pandas.DataFrame = FakeClass
pandas.Series = FakeClass



r"""Define a DFS iterator class for handling mapping data in
traversal."""

from __future__ import annotations

__all__ = ["MappingIterator"]

from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any

from coola.iterator.dfs.base import BaseIterator

if TYPE_CHECKING:
    from coola.iterator.dfs.registry import IteratorRegistry


class MappingIterator(BaseIterator[Mapping[Any, Any]]):
    r"""Iterator for depth-first traversal of mapping data structures.

    This iterator handles dict-like objects by recursively iterating over
    their values. Keys are not yielded during iteration - only the values
    are traversed. If values contain nested structures (lists, dicts, etc.),
    those are recursively iterated as well.

    Example:
        ```pycon
        >>> from coola.iterator.dfs import (
        ...     IteratorRegistry,
        ...     MappingIterator,
        ...     IterableIterator,
        ...     DefaultIterator,
        ... )
        >>> iterator = MappingIterator()
        >>> registry = IteratorRegistry(
        ...     {object: DefaultIterator(), dict: iterator, list: IterableIterator()}
        ... )
        >>> # Simple dictionary with scalar values
        >>> list(iterator.iterate({"a": 1, "b": 2}, registry))
        [1, 2]
        >>> # Nested dictionary with scalar values
        >>> list(iterator.iterate({"a": {"b": 1, "c": 2}, "d": 3}, registry))
        [1, 2, 3]
        >>> # Dictionary with list values
        >>> list(iterator.iterate({"x": [1, 2], "y": [3, 4]}, registry))
        [1, 2, 3, 4]

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def iterate(self, data: Mapping[Any, Any], registry: IteratorRegistry) -> Iterator[Any]:
        for value in data.values():
            yield from registry.iterate(value)



r"""Implement handler to check if the objects are equal."""

from __future__ import annotations

__all__ = [
    "EqualHandler",
    "EqualNanHandler",
    "SupportsEqual",
    "SupportsEqualNan",
]

import logging
from typing import TYPE_CHECKING, Protocol

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class SupportsEqual(Protocol):
    r"""Implement a protocol to represent objects with a ``equal``
    method."""

    def equal(self, other: object) -> bool:
        r"""Return ``True`` if the two objects are equal, otherwise
        ``False``.

        Args:
            other: The value to compare with.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``
        """


class SupportsEqualNan(Protocol):
    r"""Implement a protocol to represent objects with a ``equal``
    method with an option compare NaNs."""

    def equal(self, other: object, equal_nan: bool = False) -> bool:
        r"""Return ``True`` if the two objects are equal, otherwise
        ``False``.

        Args:
            other: The value to compare with.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``
        """


class EqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same data.

    This handler returns ``False`` if the two objects are different
    data, otherwise it returns ``True``. The first object must have
    a ``equal`` attribute which indicates if the two objects are
    equal or not.

    Example:
        ```pycon
        >>> import math
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import EqualHandler
        >>> class MyFloat:
        ...     def __init__(self, value: float) -> None:
        ...         self._value = float(value)
        ...     def equal(self, other: object) -> bool:
        ...         if type(other) is not type(self):
        ...             return False
        ...         return self._value == other._value
        ...
        >>> config = EqualityConfig()
        >>> handler = EqualHandler()
        >>> handler.handle(MyFloat(42), MyFloat(42), config)
        True
        >>> handler.handle(MyFloat(42), 42, config)
        False
        >>> handler.handle(MyFloat(42), float("nan"), config)
        False

        ```
    """

    def handle(self, actual: SupportsEqual, expected: object, config: EqualityConfig) -> bool:
        if not actual.equal(expected):
            if config.show_difference:
                logger.info(format_value_difference(actual=actual, expected=expected))
            return False
        return True


class EqualNanHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same data.

    This handler returns ``False`` if the two objects are different
    data, otherwise it returns ``True``. The first object must have
    a ``equal`` attribute which indicates if the two objects are
    equal or not.

    Example:
        ```pycon
        >>> import math
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import EqualNanHandler
        >>> class MyFloat:
        ...     def __init__(self, value: float) -> None:
        ...         self._value = float(value)
        ...     def equal(self, other: object, equal_nan: bool = False) -> bool:
        ...         if type(other) is not type(self):
        ...             return False
        ...         if equal_nan and math.isnan(self._value) and math.isnan(other._value):
        ...             return True
        ...         return self._value == other._value
        ...
        >>> config = EqualityConfig()
        >>> handler = EqualNanHandler()
        >>> handler.handle(MyFloat(42), MyFloat(42), config)
        True
        >>> handler.handle(MyFloat(42), 42, config)
        False
        >>> handler.handle(MyFloat(float("nan")), MyFloat(float("nan")), config)
        False
        >>> config.equal_nan = True
        >>> handler.handle(MyFloat(float("nan")), MyFloat(float("nan")), config)
        True

        ```
    """

    def handle(self, actual: SupportsEqualNan, expected: object, config: EqualityConfig) -> bool:
        if not actual.equal(expected, equal_nan=config.equal_nan):
            if config.show_difference:
                logger.info(format_value_difference(actual=actual, expected=expected))
            return False
        return True



r"""Implement handlers for ``torch.Tensor``s."""

from __future__ import annotations

__all__ = ["TorchTensorEqualHandler", "TorchTensorSameDeviceHandler"]

import logging
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    import torch

    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class TorchTensorEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two tensors are equal.

    This handler returns ``True`` if the two tensors are equal,
    otherwise ``False``. This handler is designed to be used at
    the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import torch
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import TorchTensorEqualHandler
        >>> config = EqualityConfig()
        >>> handler = TorchTensorEqualHandler()
        >>> handler.handle(torch.ones(2, 3), torch.ones(2, 3), config)
        True
        >>> handler.handle(torch.ones(2, 3), torch.zeros(2, 3), config)
        False

        ```
    """

    def handle(
        self,
        actual: torch.Tensor,
        expected: torch.Tensor,
        config: EqualityConfig,
    ) -> bool:
        object_equal = tensor_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="torch.Tensors")
            )
        return object_equal


class TorchTensorSameDeviceHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two tensors have the same device.

    This handler returns ``False`` if the two objects have different
    devices, otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> import torch
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import TrueHandler, TorchTensorSameDeviceHandler
        >>> config = EqualityConfig()
        >>> handler = TorchTensorSameDeviceHandler(next_handler=TrueHandler())
        >>> handler.handle(torch.ones(2, 3), torch.ones(3, 2), config)
        True

        ```
    """

    def handle(
        self,
        actual: torch.Tensor,
        expected: torch.Tensor,
        config: EqualityConfig,
    ) -> bool:
        if actual.device != expected.device:
            if config.show_difference:
                logger.info(
                    f"torch.Tensors have different devices: {actual.device} vs {expected.device}"
                )
            return False
        return self._handle_next(actual, expected, config=config)


def tensor_equal(tensor1: torch.Tensor, tensor2: torch.Tensor, config: EqualityConfig) -> bool:
    r"""Indicate if the two tensors are equal within a tolerance.

    Args:
        tensor1: The first tensor to compare.
        tensor2: The second tensor to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two tensors are equal within a tolerance,
            otherwise ``False``.
    """
    if config.equal_nan or config.atol > 0 or config.rtol > 0:
        return tensor1.allclose(
            tensor2, atol=config.atol, rtol=config.rtol, equal_nan=config.equal_nan
        )
    return tensor1.equal(tensor2)



r"""Implement handler to check if the objects have the same data
type."""

from __future__ import annotations

__all__ = ["SameDTypeHandler", "SupportsDType"]

import logging
from typing import TYPE_CHECKING, Any, Protocol

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig


logger: logging.Logger = logging.getLogger(__name__)


class SupportsDType(Protocol):
    r"""Implement a protocol to represent objects with a ``dtype``
    attribute.

    This protocol can be used to represent several objects like
    ``jax.numpy.ndarray``s, ``numpy.ndarray``s,  and
    ``torch.Tensor``s.
    """

    @property
    def dtype(self) -> Any:
        return  # pragma: no cover


class SameDTypeHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same data type.

    This handler returns ``False`` if the two objects have different
    data types, otherwise it passes the inputs to the next handler.
    The objects must have a ``dtype`` attribute (e.g. ``object.dtype``)
    which returns the data type of the object. This handler works on
    ``numpy.ndarray``s and ``torch.Tensor``s objects.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameDTypeHandler, TrueHandler
        >>> config = EqualityConfig()
        >>> handler = SameDTypeHandler(next_handler=TrueHandler())
        >>> handler.handle(np.ones((2, 3)), np.ones((2, 3)), config)
        True
        >>> handler.handle(np.ones((2, 3), dtype=float), np.ones((2, 3), dtype=int), config)
        False

        ```
    """

    def handle(
        self, actual: SupportsDType, expected: SupportsDType, config: EqualityConfig
    ) -> bool:
        if actual.dtype != expected.dtype:
            if config.show_difference:
                logger.info(
                    f"objects have different data types: {actual.dtype} vs {expected.dtype}"
                )
            return False
        return self._handle_next(actual, expected, config=config)



r"""Define the public interface to summarize nested data."""

from __future__ import annotations

__all__ = ["get_default_registry", "register_summarizers", "summarize"]

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any

from coola.summary.default import DefaultSummarizer
from coola.summary.mapping import MappingSummarizer
from coola.summary.numpy import NDArraySummarizer
from coola.summary.registry import SummarizerRegistry
from coola.summary.sequence import SequenceSummarizer
from coola.summary.set import SetSummarizer
from coola.summary.torch import TensorSummarizer
from coola.utils.imports import is_numpy_available, is_torch_available

if TYPE_CHECKING:
    from coola.summary.base import BaseSummarizer

if is_torch_available():  # pragma: no cover
    import torch

if is_numpy_available():  # pragma: no cover
    import numpy as np


def summarize(data: object, max_depth: int = 1, registry: SummarizerRegistry | None = None) -> str:
    r"""Create a summary string representation of nested data.

    Args:
        data: Input data (can be nested)
        max_depth: The maximum nesting level to expand when summarizing.
            Structures deeper than this level are shown in compact form.
            Must be non-negative. Default is 1, which expands only the
            top level of nested structures.
        registry: Registry to resolve summarizers for nested data.
            If None, uses the default registry.

    Returns:
        String summary of the data

    Example:
        ```pycon
        >>> from coola.summary import summarize
        >>> print(summarize({"a": 1, "b": "abc"}))
        <class 'dict'> (length=2)
          (a): 1
          (b): abc

        ```
    """
    if registry is None:
        registry = get_default_registry()
    return registry.summarize(data, max_depth=max_depth)


def register_summarizers(
    mapping: Mapping[type, BaseSummarizer[Any]],
    exist_ok: bool = False,
) -> None:
    """Register custom summarizers to the default global registry.

    This allows users to add support for custom types without modifying
    global state directly.

    Args:
        mapping: Dictionary mapping types to summarizer instances
        exist_ok: If False, raises error if any type already registered

    Example:
        ```pycon
        >>> from coola.summary import register_summarizers, BaseSummarizer, SummarizerRegistry
        >>> class MyType:
        ...     def __init__(self, value):
        ...         self.value = value
        ...
        >>> class MySummarizer(BaseSummarizer[MyType]):
        ...     def equal(self, other: object) -> bool:
        ...         return type(object) is type(self)
        ...     def summarize(
        ...         self,
        ...         data: MyType,
        ...         registry: SummarizerRegistry,
        ...         depth: int = 0,
        ...         max_depth: int = 1,
        ...     ) -> str:
        ...         return f"<MyType> value={data.value}"
        ...
        >>> register_summarizers({MyType: MySummarizer()})

        ```
    """
    get_default_registry().register_many(mapping, exist_ok=exist_ok)


def get_default_registry() -> SummarizerRegistry:
    """Get or create the default global registry with common Python
    types.

    Returns a singleton registry instance that is pre-configured with summarizers
    for Python's built-in types including sequences (list, tuple), mappings (dict),
    sets, and scalar types (int, float, str, bool).

    This function uses a singleton pattern to ensure the same registry instance
    is returned on subsequent calls, which is efficient and maintains consistency
    across an application.

    Returns:
        A SummarizerRegistry instance with summarizers registered for:
            - Scalar types (int, float, complex, bool, str)
            - Sequences (list, tuple, Sequence ABC)
            - Sets (set, frozenset)
            - Mappings (dict, Mapping ABC)

    Notes:
        The singleton pattern means modifications to the returned registry
        affect all future calls to this function. If you need an isolated
        registry, create a new SummarizerRegistry instance directly.

    Example:
        ```pycon
        >>> from coola.summary import get_default_registry
        >>> registry = get_default_registry()
        >>> # Registry is ready to use with common Python types
        >>> print(registry.summarize([1, 2, 3]))
        <class 'list'> (length=3)
          (0): 1
          (1): 2
          (2): 3
        >>> print(registry.summarize({"a": 1, "b": 2}))
        <class 'dict'> (length=2)
          (a): 1
          (b): 2

        ```
    """
    if not hasattr(get_default_registry, "_registry"):
        registry = SummarizerRegistry()
        _register_default_summarizers(registry)
        get_default_registry._registry = registry
    return get_default_registry._registry


def _register_default_summarizers(registry: SummarizerRegistry) -> None:
    """Register default summarizers for common Python types.

    This internal function sets up the standard type-to-summarizer mappings
    used by the default registry. It registers summarizers in a specific order
    to ensure proper inheritance handling via MRO.

    The registration strategy:
        - Scalar types use DefaultSummarizer (no recursion)
        - Strings use DefaultSummarizer to prevent character-by-character iteration
        - Sequences use SequenceSummarizer for summary list/tuple processing
        - Sets use SetSummarizer for summary set processing
        - Mappings use MappingSummarizer for summary dict processing

    Args:
        registry: The registry to populate with default summarizers

    Notes:
        This function is called internally by get_default_registry() and should
        not typically be called directly by users.
    """
    summarizers: dict[type, BaseSummarizer[Any]] = _get_native_summarizers()
    summarizers.update(_get_numpy_summarizers())
    summarizers.update(_get_torch_summarizers())
    registry.register_many(summarizers)


def _get_native_summarizers() -> dict[type, BaseSummarizer[Any]]:
    r"""Get the native summarizers for common Python types.

    Returns:
        The native summarizers for common Python types.

    Notes:
        This function is called internally by get_default_registry() and should
        not typically be called directly by users.
    """
    default_summarizer = DefaultSummarizer()
    sequence_summarizer = SequenceSummarizer()
    set_summarizer = SetSummarizer()
    mapping_summarizer = MappingSummarizer()
    return {
        # Object is the catch-all base for unregistered types
        object: default_summarizer,
        # Strings should not be iterated character by character
        str: default_summarizer,
        # Numeric types - no recursion needed
        int: default_summarizer,
        float: default_summarizer,
        complex: default_summarizer,
        bool: default_summarizer,
        # Sequences - summary transformation preserving order
        list: sequence_summarizer,
        tuple: sequence_summarizer,
        Sequence: sequence_summarizer,
        # Sets - summary transformation without order
        set: set_summarizer,
        frozenset: set_summarizer,
        # Mappings - summary transformation of keys and values
        dict: mapping_summarizer,
        Mapping: mapping_summarizer,
    }


def _get_numpy_summarizers() -> dict[type, BaseSummarizer[Any]]:
    r"""Get the native summarizers for NumPy types.

    Returns:
        The summarizers for NumPy array types.

    Notes:
        This function is called internally by get_default_registry() and should
        not typically be called directly by users.
    """
    if not is_numpy_available():  # pragma: no cover
        return {}
    return {np.ndarray: NDArraySummarizer()}


def _get_torch_summarizers() -> dict[type, BaseSummarizer[Any]]:
    r"""Get the native summarizers for PyTorch types.

    Returns:
        The summarizers for PyTorch tensor types.

    Notes:
        This function is called internally by get_default_registry() and should
        not typically be called directly by users.
    """
    if not is_torch_available():  # pragma: no cover
        return {}
    return {torch.Tensor: TensorSummarizer()}



r"""Implement a reducer using NumPy functions.

The reducer is registered only if ``numpy`` is available.
"""

from __future__ import annotations

__all__ = ["NumpyReducer"]

from collections.abc import Sequence
from typing import TypeVar

from coola.reducer.base import BaseBasicReducer
from coola.utils.imports import check_numpy, is_numpy_available

if is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    from coola.utils.fallback.numpy import numpy as np


T = TypeVar("T", Sequence[float], np.ndarray)


class NumpyReducer(BaseBasicReducer[T]):
    r"""Implement a reducer based on NumPy functions.

    Raises:
        RuntimeError: if ``numpy`` is not installed.

    Example:
        ```pycon
        >>> from coola.reducer import NumpyReducer
        >>> reducer = NumpyReducer()
        >>> reducer.max([-2, -1, 0, 1, 2])
        2
        >>> reducer.median([-2, -1, 0, 1, 2])
        0.0
        >>> reducer.sort([2, 1, -2, 3, 0])
        [-2, 0, 1, 2, 3]

        ```
    """

    def __init__(self) -> None:
        check_numpy()

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _is_empty(self, values: T) -> bool:
        if isinstance(values, np.ndarray):
            return values.size == 0
        return not values

    def _max(self, values: T) -> int | float:
        return np.max(np.asarray(values)).item()

    def _mean(self, values: T) -> float:
        return np.mean(np.asarray(values)).item()

    def _median(self, values: T) -> int | float:
        return np.median(np.asarray(values)).item()

    def _min(self, values: T) -> int | float:
        return np.min(np.asarray(values)).item()

    def _quantile(self, values: T, quantiles: Sequence[float]) -> list[float]:
        return np.quantile(np.asarray(values), q=quantiles).tolist()

    def sort(self, values: T, descending: bool = False) -> list[int | float]:
        array = np.sort(np.asarray(values))
        if descending:
            return array[::-1].tolist()
        return array.tolist()

    def _std(self, values: T) -> float:
        if len(values) <= 1:
            return float("nan")
        return np.std(np.asarray(values), ddof=1).item()



r"""Implement handlers for ``pyarrow`` objects."""

from __future__ import annotations

__all__ = ["PyarrowEqualHandler"]

import logging
import warnings
from contextlib import suppress
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    import pyarrow as pa

    from coola.equality.config import EqualityConfig


logger: logging.Logger = logging.getLogger(__name__)


class PyarrowEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two pyarrow arrays or tables are equal.

    This handler returns ``True`` if the two arrays or tables are
    equal, otherwise ``False``. This handler is designed to be used at
    the end of the chain of responsibility. This handler does
    not call the next handler.

    Note that ``config.equal_nan``, ``config.atol`` and ``config.rtol``
    arguments are ignored.

    Example:
        ```pycon
        >>> import pyarrow
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PyarrowEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PyarrowEqualHandler()
        >>> handler.handle(pyarrow.array([1, 2, 3]), pyarrow.array([1, 2, 3]), config)
        True
        >>> handler.handle(pyarrow.array([1, 2, 3]), pyarrow.array([1, 2, 4]), config)
        False

        ```
    """

    def handle(
        self,
        actual: pa.Array | pa.Table,
        expected: pa.Array | pa.Table,
        config: EqualityConfig,
    ) -> bool:
        equal = object_equal(actual, expected, config)
        if config.show_difference and not equal:
            logger.info(format_value_difference(actual=actual, expected=expected))
        return equal


def object_equal(
    obj1: pa.Array | pa.Table, obj2: pa.Array | pa.Table, config: EqualityConfig
) -> bool:
    r"""Indicate if the two arrays or tables are equal within a
    tolerance.

    ``config.equal_nan``, ``config.atol`` and ``config.rtol``
    arguments are ignored.

    Args:
        obj1: The first array or table to compare.
        obj2: The second array or table to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two arrays or tables are equal within a
            tolerance, otherwise ``False``.

    Example:
        ```pycon
        >>> import pyarrow as pa
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler.pyarrow import object_equal
        >>> config = EqualityConfig()
        >>> object_equal(pa.array([1, 2, 3]), pa.array([1, 2, 3]), config)
        True
        >>> object_equal(pa.array([1, 2, 3]), pa.array([1, 2, 4]), config)
        False

        ```
    """
    if config.equal_nan:
        warnings.warn(
            f"equal_nan is ignored because it is not supported for {type(obj1)}",
            RuntimeWarning,
            stacklevel=3,
        )
    if config.atol > 0:
        warnings.warn(
            f"atol is ignored because it is not supported for {type(obj1)}",
            RuntimeWarning,
            stacklevel=3,
        )
    if config.rtol > 0:
        warnings.warn(
            f"rtol is ignored because it is not supported for {type(obj1)}",
            RuntimeWarning,
            stacklevel=3,
        )
    with suppress(TypeError):
        return obj1.equals(obj2)
    return False



r"""Define the equality handler base classes."""

from __future__ import annotations

__all__ = ["HandlerEqualityMixin"]

from typing import TYPE_CHECKING

from coola.equality.handler.utils import handlers_are_equal

if TYPE_CHECKING:
    from coola.equality.handler.base import BaseEqualityHandler


class HandlerEqualityMixin:
    r"""Mixin providing a standard implementation of the equal() method.

    This mixin eliminates code duplication across handlers that only need
    to compare their type and next_handler. Handlers using this mixin must
    inherit from BaseEqualityHandler to ensure the next_handler attribute
    is available.

    Design Note:
        This mixin must be used with classes that inherit from
        ``BaseEqualityHandler``. The type annotation ``self: BaseEqualityHandler``
        on the ``equal()`` method enforces this constraint and enables
        type-safe access to the ``next_handler`` attribute.

    Example:
        ```pycon
        >>> from coola.equality.handler import BaseEqualityHandler, HandlerEqualityMixin
        >>> class MyHandler(HandlerEqualityMixin, BaseEqualityHandler):
        ...     def handle(self, actual, expected, config):
        ...         return True
        ...
        >>> handler1 = MyHandler()
        >>> handler2 = MyHandler()
        >>> handler1.equal(handler2)
        True

        ```
    """

    def equal(self: BaseEqualityHandler, other: object) -> bool:
        r"""Indicate if two handlers are equal.

        Two handlers are equal if they are of the same type and have
        equal next_handler chains.

        Note:
            The type annotation ``self: BaseEqualityHandler`` ensures this
            mixin is only used with BaseEqualityHandler subclasses, enabling
            type-safe access to ``next_handler``.

        Args:
            other: The other object to compare with.

        Returns:
            ``True`` if the handlers are equal, otherwise ``False``.
        """
        if type(other) is not type(self):
            return False
        return handlers_are_equal(self.next_handler, other.next_handler)



r"""Contain fallback implementations used when ``torch`` dependency is
not available."""

from __future__ import annotations

__all__ = ["cuda", "nn", "torch"]

from types import ModuleType
from typing import Any, NoReturn

from coola.utils.imports import raise_error_torch_missing


class FakeClass:
    r"""Fake class that raises an error because torch is not installed.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Raises:
        RuntimeError: torch is required for this functionality.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ARG002
        raise_error_torch_missing()


def fake_function(*args: Any, **kwargs: Any) -> NoReturn:  # noqa: ARG001
    r"""Fake function that raises an error because torch is not
    installed.

    Args:
        *args: Positional arguments.
        **kwargs: Keyword arguments.

    Raises:
        RuntimeError: torch is required for this functionality.
    """
    raise_error_torch_missing()


cuda: ModuleType = ModuleType("torch.cuda")
cuda.is_available = fake_function
cuda.synchronize = fake_function

nn: ModuleType = ModuleType("torch.nn")
nn.utils = ModuleType("torch.nn.utils")
nn.utils.rnn = ModuleType("torch.nn.utils.rnn")
nn.utils.rnn.PackedSequence = FakeClass

# Create a fake torch package
torch: ModuleType = ModuleType("torch")
torch.cuda = cuda
torch.nn = nn

torch.Tensor = FakeClass
torch.tensor = fake_function



r"""Contain path utility functions."""

from __future__ import annotations

__all__ = ["sanitize_path", "working_directory"]

import contextlib
import os
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import unquote, urlparse

if TYPE_CHECKING:
    from collections.abc import Generator


def sanitize_path(path: Path | str) -> Path:
    r"""Sanitize the given path.

    Args:
        path: The path to sanitize. The path can be a string or a
            ``pathlib.Path`` object.

    Returns:
        The sanitized path as a ``pathlib.Path`` object.

    Example:
        ```pycon
        >>> from pathlib import Path
        >>> from coola.utils.path import sanitize_path
        >>> sanitize_path("something")
        PosixPath('.../something')
        >>> sanitize_path("")
        PosixPath('...')
        >>> sanitize_path(Path("something"))
        PosixPath('.../something')
        >>> sanitize_path(Path("something/./../"))
        PosixPath('...')

        ```
    """
    if isinstance(path, str):
        # use urlparse to parse file URI
        # source: https://stackoverflow.com/a/15048213
        path = Path(unquote(urlparse(path).path))
    return path.expanduser().resolve()


@contextlib.contextmanager
def working_directory(path: Path) -> Generator[None]:
    r"""Context manager to change the working directory to the given
    path, and then changes it back to its previous value on exit.

    source: https://gist.github.com/nottrobin/3d675653244f8814838a

    Args:
        path: The path to the temporary working directory.

    Example:
        ```pycon
        >>> from coola.utils.path import working_directory
        >>> with working_directory(Path("src")):
        ...     x = 1
        ...

        ```
    """
    path = sanitize_path(path)
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)



r"""Implement handlers for ``pandas.DataFrame``s and
``pandas.Series``s."""

from __future__ import annotations

__all__ = ["PandasDataFrameEqualHandler", "PandasSeriesEqualHandler"]

import logging
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.utils.imports import is_pandas_available

if TYPE_CHECKING or is_pandas_available():
    import pandas as pd
else:  # pragma: no cover
    from coola.utils.fallback.pandas import pandas as pd

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class PandasDataFrameEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two ``pandas.DataFrame``s are equal.

    This handler returns ``True`` if the two ``pandas.DataFrame``s are
    equal, otherwise ``False``. This handler is designed to be used
    at the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import pandas
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PandasDataFrameEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PandasDataFrameEqualHandler()
        >>> handler.handle(
        ...     pandas.DataFrame({"col": [1, 2, 3]}),
        ...     pandas.DataFrame({"col": [1, 2, 3]}),
        ...     config,
        ... )
        True
        >>> handler.handle(
        ...     pandas.DataFrame({"col": [1, 2, 3]}),
        ...     pandas.DataFrame({"col": [1, 2, 4]}),
        ...     config,
        ... )
        False

        ```
    """

    def handle(
        self,
        actual: pd.DataFrame,
        expected: pd.DataFrame,
        config: EqualityConfig,
    ) -> bool:
        object_equal = frame_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="pandas.DataFrames")
            )
        return object_equal


class PandasSeriesEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two ``pandas.Series``es are equal.

    This handler returns ``True`` if the two ``pandas.Series``es are
    equal, otherwise ``False``. This handler is designed to be used
    at the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import pandas
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import PandasSeriesEqualHandler
        >>> config = EqualityConfig()
        >>> handler = PandasSeriesEqualHandler()
        >>> handler.handle(pandas.Series([1, 2, 3]), pandas.Series([1, 2, 3]), config)
        True
        >>> handler.handle(pandas.Series([1, 2, 3]), pandas.Series([1, 2, 4]), config)
        False

        ```
    """

    def handle(
        self,
        actual: pd.Series,
        expected: pd.Series,
        config: EqualityConfig,
    ) -> bool:
        object_equal = series_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(
                format_value_difference(actual=actual, expected=expected, name="pandas.Series")
            )
        return object_equal


def frame_equal(df1: pd.DataFrame, df2: pd.DataFrame, config: EqualityConfig) -> bool:
    r"""Indicate if the two DataFrames are equal or not.

    Args:
        df1: The first DataFrame to compare.
        df2: The second DataFrame to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two DataFrames are equal, otherwise ``False``.
    """
    if not config.equal_nan and df1.isna().any().any():
        return False
    try:
        pd.testing.assert_frame_equal(
            df1,
            df2,
            check_exact=config.atol == 0 and config.rtol == 0,
            check_index_type=True,
            atol=config.atol,
            rtol=config.rtol,
        )
    except AssertionError:
        return False
    return True


def series_equal(series1: pd.Series, series2: pd.Series, config: EqualityConfig) -> bool:
    r"""Indicate if the two series are equal or not.

    Args:
        series1: The first series to compare.
        series2: The second series to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two series are equal, otherwise ``False``.
    """
    if not config.equal_nan and series1.isna().any():
        return False
    try:
        pd.testing.assert_series_equal(
            series1,
            series2,
            check_exact=config.atol == 0 and config.rtol == 0,
            check_index_type=True,
            atol=config.atol,
            rtol=config.rtol,
        )
    except AssertionError:
        return False
    return True



r"""Type-based filtering utilities for iterators.

This module provides utilities for filtering iterators based on type
checks, allowing you to extract only values matching specific types from
heterogeneous sequences.
"""

from __future__ import annotations

__all__ = ["filter_by_type"]

from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

T = TypeVar("T")


def filter_by_type(iterator: Iterable[Any], types: type[T] | tuple[type, ...]) -> Iterator[T]:
    """Filter an iterator to yield only values of specified types.

    This function acts as a type-safe filter that passes through only values
    matching the specified type(s). It's particularly useful when working with
    heterogeneous collections where you need to extract elements of specific
    types.

    Args:
        iterator: An iterator or iterable to filter. Can contain values of any type.
        types: A single type or tuple of types to filter for. Only values that are
            instances of these types will be yielded. Follows the same semantics
            as the built-in isinstance() function.

    Yields:
        Values from the input iterator that are instances of any of the specified
        types, preserving the original order of elements.

    Example:
        Filter mixed-type list to get only integers:

        ```pycon
        >>> from coola.iterator import filter_by_type
        >>> list(filter_by_type([1, "hello", 2, 3.14, "world", 4], int))
        [1, 2, 4]

        ```

        Filter for multiple types:

        ```pycon
        >>> from coola.iterator import filter_by_type
        >>> # Note: bool is a subclass of int
        >>> list(filter_by_type([1, "hello", 2.5, True, None, [1, 2]], (int, float)))
        [1, 2.5, True]

        ```

    Notes:
        - This function uses isinstance() internally, so subclass relationships
          are respected (e.g., bool values will match int type).
        - The input iterator is consumed as items are yielded.
        - For empty iterators or when no items match, the generator yields nothing.
    """
    for value in iterator:
        if isinstance(value, types):
            yield value



r"""Implement equality testers for PyArrow arrays and tables.

This module provides equality testers for pyarrow.Array and
pyarrow.Table using PyArrow's built-in equals() method.
"""

from __future__ import annotations

__all__ = ["PyarrowEqualityTester"]

from typing import TYPE_CHECKING

from coola.equality.handler import (
    PyarrowEqualHandler,
    SameObjectHandler,
    SameTypeHandler,
    create_chain,
)
from coola.equality.tester.base import BaseEqualityTester
from coola.utils.imports import check_pyarrow, is_pyarrow_available

if TYPE_CHECKING or is_pyarrow_available():
    import pyarrow as pa
else:  # pragma: no cover
    from coola.utils.fallback.pyarrow import pyarrow as pa

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig


class PyarrowEqualityTester(BaseEqualityTester[pa.Array]):
    r"""Implement an equality tester for ``pyarrow.Array``s and
    ``pyarrow.Table``s.

    This tester uses PyArrow's equals() method for comparison. The handler chain:
    1. SameObjectHandler: Check for object identity
    2. SameTypeHandler: Verify both are pyarrow objects
    3. PyarrowEqualHandler: Use PyArrow's equals() method

    Note:
        The ``config.equal_nan``, ``config.atol``, and ``config.rtol`` arguments
        are ignored as PyArrow's equals() method does not support these parameters.
        PyArrow performs its own equality checking logic.

    Example:
        Basic array comparison:

        ```pycon
        >>> import pyarrow as pa
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.tester import PyarrowEqualityTester
        >>> config = EqualityConfig()
        >>> tester = PyarrowEqualityTester()
        >>> tester.objects_are_equal(pa.array([1, 2, 3]), pa.array([1, 2, 3]), config)
        True
        >>> tester.objects_are_equal(pa.array([1, 2, 3]), pa.array([1, 2, 4]), config)
        False

        ```

        Table comparison:

        ```pycon
        >>> import pyarrow as pa
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.tester import PyarrowEqualityTester
        >>> config = EqualityConfig()
        >>> tester = PyarrowEqualityTester()
        >>> table1 = pa.table({"col": [1, 2, 3]})
        >>> table2 = pa.table({"col": [1, 2, 3]})
        >>> tester.objects_are_equal(table1, table2, config)
        True

        ```
    """

    def __init__(self) -> None:
        check_pyarrow()
        self._handler = create_chain(SameObjectHandler(), SameTypeHandler(), PyarrowEqualHandler())

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def equal(self, other: object) -> bool:
        return type(other) is type(self)

    def objects_are_equal(
        self,
        actual: pa.Array,
        expected: object,
        config: EqualityConfig,
    ) -> bool:
        return self._handler.handle(actual, expected, config=config)



r"""Implement handlers for mapping objects."""

from __future__ import annotations

__all__ = ["MappingSameKeysHandler", "MappingSameValuesHandler"]

import logging
from typing import TYPE_CHECKING, Any

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_mapping_difference
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.equality.handler.utils import check_recursion_depth

if TYPE_CHECKING:
    from collections.abc import Mapping

    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class MappingSameKeysHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same keys.

    This handler returns ``False`` if the two objects have different
    keys, otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import MappingSameKeysHandler
        >>> config = EqualityConfig()
        >>> handler = MappingSameKeysHandler()
        >>> handler.handle({"a": 1, "b": 2}, {"a": 1, "b": 2, "c": 1}, config)
        False

        ```
    """

    def handle(
        self,
        actual: Mapping[Any, Any],
        expected: Mapping[Any, Any],
        config: EqualityConfig,
    ) -> bool:
        keys1 = set(actual.keys())
        keys2 = set(expected.keys())
        if keys1 != keys2:
            if config.show_difference:
                logger.info(
                    format_mapping_difference(
                        missing_keys=keys1 - keys2, additional_keys=keys2 - keys1
                    )
                )
            return False
        return self._handle_next(actual, expected, config=config)


class MappingSameValuesHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the key-value pairs in the first mapping are in the
    second mapping.

    This handler returns ``False`` if the one of the key-value pair in
    the first mapping is not in the second mapping, otherwise it
    passes the inputs to the next handler.

    Notes:
        This handler assumes that all the keys in the first mapping are
        also in the second mapping. The second mapping can have more
        keys. To check if two mappings are equal, you can combine this
        handler with ``MappingSameKeysHandler``.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import MappingSameValuesHandler, TrueHandler
        >>> config = EqualityConfig()
        >>> handler = MappingSameValuesHandler(next_handler=TrueHandler())
        >>> handler.handle({"a": 1, "b": 2}, {"a": 1, "b": 2}, config)
        True
        >>> handler.handle({"a": 1, "b": 2}, {"a": 1, "b": 3}, config)
        False

        ```
    """

    def handle(
        self,
        actual: Mapping[Any, Any],
        expected: Mapping[Any, Any],
        config: EqualityConfig,
    ) -> bool:
        with check_recursion_depth(config):
            for key in actual:
                if not config.registry.objects_are_equal(actual[key], expected[key], config):
                    if config.show_difference:
                        logger.info(format_mapping_difference(different_value_key=key))
                    return False
            return self._handle_next(actual, expected, config=config)



r"""Implement handlers to check if the objects are equal."""

from __future__ import annotations

__all__ = ["NanEqualHandler", "ScalarEqualHandler"]

import logging
import math
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig


logger: logging.Logger = logging.getLogger(__name__)


class NanEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two NaNs are equal.

    This handler returns ``True`` if the two numbers are NaNs,
    otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import NanEqualHandler, FalseHandler
        >>> config = EqualityConfig()
        >>> handler = NanEqualHandler(next_handler=FalseHandler())
        >>> handler.handle(float("nan"), float("nan"), config)
        False
        >>> config.equal_nan = True
        >>> handler.handle(float("nan"), float("nan"), config)
        True

        ```
    """

    def handle(
        self,
        actual: float,
        expected: float,
        config: EqualityConfig,
    ) -> bool:
        if config.equal_nan and math.isnan(actual) and math.isnan(expected):
            return True
        return self._handle_next(actual, expected, config=config)


class ScalarEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two numbers are equal or not.

    This handler returns ``False`` if the two numbers are
    different, otherwise it returns ``True``. It is possible to
    control the tolerance by using ``atol`` and ``rtol``.
    By default, the tolerances are set to 0.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import ScalarEqualHandler
        >>> config = EqualityConfig()
        >>> handler = ScalarEqualHandler()
        >>> handler.handle(42.0, 42.0, config)
        True
        >>> config.atol = 1e-3
        >>> handler.handle(42.0, 42.0001, config)
        True
        >>> handler.handle(float("nan"), float("nan"), config)
        False

        ```
    """

    def handle(self, actual: float, expected: float, config: EqualityConfig) -> bool:
        object_equal = number_equal(actual, expected, config)
        if not object_equal and config.show_difference:
            logger.info(format_value_difference(actual=actual, expected=expected, name="numbers"))
        return object_equal


def number_equal(number1: float, number2: float, config: EqualityConfig) -> bool:
    r"""Indicate if the two numbers are equal within a tolerance.

    Args:
        number1: The first number to compare.
        number2: The second number to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two numbers are equal within a tolerance,
            otherwise ``False``.
    """
    if config.atol > 0.0 or config.rtol > 0.0:
        return math.isclose(number1, number2, abs_tol=config.atol, rel_tol=config.rtol)
    return number1 == number2



r"""Define the equality tester registry for recursive equality
comparison.

This module provides a registry system that manages and dispatches
equality testers based on data types, enabling recursive equality
checking of nested data structures.
"""

from __future__ import annotations

__all__ = ["EqualityTesterRegistry"]

from typing import TYPE_CHECKING, Any

from coola.equality.tester.base import BaseEqualityTester
from coola.utils.format import repr_indent, repr_mapping, str_indent, str_mapping

if TYPE_CHECKING:
    from collections.abc import Mapping

    from coola.equality.config import EqualityConfig


class EqualityTesterRegistry:
    """Registry that manages and dispatches equality testers based on
    data type.

    This registry maintains a mapping from Python types to equality tester instances
    and uses the Method Resolution Order (MRO) for type lookup. When checking
    equality, it automatically selects the most specific registered equality tester for
    the data's type, falling back to parent types or a default tester if needed.

    The registry includes an LRU cache for type lookups to optimize performance
    in applications that repeatedly check equality of similar data structures.

    Args:
        initial_state: Optional initial mapping of types to equality testers.
            If provided, the state is copied to prevent external mutations.

    Attributes:
        _state: Internal mapping of registered types to equality testers

    Example:
        Basic usage:

        ```pycon
        >>> from coola.equality.tester import (
        ...     EqualityTesterRegistry,
        ...     SequenceEqualityTester,
        ...     DefaultEqualityTester,
        ... )
        >>> from coola.equality.config import EqualityConfig
        >>> registry = EqualityTesterRegistry(
        ...     {object: DefaultEqualityTester(), list: SequenceEqualityTester()}
        ... )
        >>> registry
        EqualityTesterRegistry(
          (state): TypeRegistry(
              (<class 'object'>): DefaultEqualityTester()
              (<class 'list'>): SequenceEqualityTester()
            )
        )
        >>> config = EqualityConfig()
        >>> registry.objects_are_equal([1, 2, 3], [1, 2, 3], config=config)
        True

        ```
    """

    def __init__(self, initial_state: dict[type, BaseEqualityTester[Any]] | None = None) -> None:
        # local import to avoid cyclic imports as TypeRegistry uses objects_are_equal
        from coola.registry.type import TypeRegistry  # noqa: PLC0415

        self._state: TypeRegistry[BaseEqualityTester] = TypeRegistry[BaseEqualityTester](
            initial_state
        )

    def __repr__(self) -> str:
        state = repr_indent(repr_mapping({"state": self._state}))
        return f"{self.__class__.__qualname__}(\n  {state}\n)"

    def __str__(self) -> str:
        state = str_indent(str_mapping({"state": self._state}))
        return f"{self.__class__.__qualname__}(\n  {state}\n)"

    def register(
        self,
        data_type: type,
        tester: BaseEqualityTester[Any],
        exist_ok: bool = False,
    ) -> None:
        """Register an equality tester for a given data type.

        This method associates an equality tester instance with a specific Python type.
        When checking equality for data of this type, the registered tester will be used.
        The cache is automatically cleared after registration to ensure consistency.

        Args:
            data_type: The Python type to register (e.g., list, dict, custom classes)
            tester: The equality tester instance that handles this type
            exist_ok: If False (default), raises an error if the type is already
                registered. If True, overwrites the existing registration silently.

        Raises:
            RuntimeError: If the type is already registered and exist_ok is False

        Example:
            ```pycon
            >>> from coola.equality.tester import EqualityTesterRegistry, SequenceEqualityTester
            >>> registry = EqualityTesterRegistry()
            >>> registry.register(list, SequenceEqualityTester())
            >>> registry.has_equality_tester(list)
            True

            ```
        """
        self._state.register(data_type, tester, exist_ok=exist_ok)

    def register_many(
        self,
        mapping: Mapping[type, BaseEqualityTester[Any]],
        exist_ok: bool = False,
    ) -> None:
        """Register multiple equality testers at once.

        This is a convenience method for bulk registration that internally calls
        register() for each type-tester pair.

        Args:
            mapping: Dictionary mapping Python types to equality tester instances
            exist_ok: If False (default), raises an error if any type is already
                registered. If True, overwrites existing registrations silently.

        Raises:
            RuntimeError: If any type is already registered and exist_ok is False

        Example:
            ```pycon
            >>> from coola.equality.tester import (
            ...     EqualityTesterRegistry,
            ...     SequenceEqualityTester,
            ...     MappingEqualityTester,
            ... )
            >>> registry = EqualityTesterRegistry()
            >>> registry.register_many(
            ...     {
            ...         list: SequenceEqualityTester(),
            ...         dict: MappingEqualityTester(),
            ...     }
            ... )
            >>> registry
            EqualityTesterRegistry(
              (state): TypeRegistry(
                  (<class 'list'>): SequenceEqualityTester()
                  (<class 'dict'>): MappingEqualityTester()
                )
            )

            ```
        """
        self._state.register_many(mapping, exist_ok=exist_ok)

    def has_equality_tester(self, data_type: type) -> bool:
        """Check if an equality tester is explicitly registered for the
        given type.

        Note that this only checks for direct registration. Even if this returns
        False, find_equality_tester() may still return a tester via MRO lookup
        or the default tester.

        Args:
            data_type: The type to check

        Returns:
            True if an equality tester is explicitly registered for this type,
            False otherwise

        Example:
            ```pycon
            >>> from coola.equality.tester import EqualityTesterRegistry, SequenceEqualityTester
            >>> registry = EqualityTesterRegistry()
            >>> registry.register(list, SequenceEqualityTester())
            >>> registry.has_equality_tester(list)
            True
            >>> registry.has_equality_tester(tuple)
            False

            ```
        """
        return data_type in self._state

    def find_equality_tester(self, data_type: type) -> BaseEqualityTester[Any]:
        """Find the appropriate equality tester for a given type.

        Uses the Method Resolution Order (MRO) to find the most specific
        registered equality tester. For example, if you register a tester
        for Sequence but not for list, lists will use the Sequence tester.

        Results are cached using an LRU cache (256 entries) for performance,
        as tester lookup is a hot path in recursive equality checking.

        Args:
            data_type: The Python type to find an equality tester for

        Returns:
            The most specific registered equality tester for this type, a parent
            type's tester via MRO, or the default tester

        Example:
            ```pycon
            >>> from collections.abc import Sequence
            >>> from coola.equality.tester import (
            ...     EqualityTesterRegistry,
            ...     SequenceEqualityTester,
            ...     DefaultEqualityTester,
            ... )
            >>> registry = EqualityTesterRegistry({object: DefaultEqualityTester()})
            >>> registry.register(Sequence, SequenceEqualityTester())
            >>> # list does not inherit from Sequence, so it uses DefaultEqualityTester
            >>> tester = registry.find_equality_tester(list)
            >>> tester
            DefaultEqualityTester()

            ```
        """
        return self._state.resolve(data_type)

    def objects_are_equal(self, actual: object, expected: object, config: EqualityConfig) -> bool:
        """Check if two objects are equal by recursively comparing their
        structure.

        This is the main entry point for equality checking. It automatically:
        1. Determines the actual object's type
        2. Finds the appropriate equality tester
        3. Delegates to that tester's objects_are_equal method
        4. The tester recursively processes nested structures

        Args:
            actual: The actual object.
            expected: The expected object.
            config: The equality configuration.

        Returns:
            True if the objects are equal according to the registered testers,
            False otherwise

        Example:
            Checking if two lists of integers are equal:

            ```pycon
            >>> from coola.equality.config import EqualityConfig
            >>> from coola.equality.tester import get_default_registry
            >>> registry = get_default_registry()
            >>> config = EqualityConfig()
            >>> registry.objects_are_equal([1, 2, 3], [1, 2, 3], config=config)
            True

            ```

            Checking if two lists of tensors are equal:

            ```pycon
            >>> import torch
            >>> from coola.equality.config import EqualityConfig
            >>> from coola.equality.tester import get_default_registry
            >>> registry = get_default_registry()
            >>> config = EqualityConfig()
            >>> registry.objects_are_equal(
            ...     [torch.ones(2, 3), torch.zeros(2)],
            ...     [torch.ones(2, 3), torch.zeros(2)],
            ...     config=config,
            ... )
            True

            ```
        """
        tester = self.find_equality_tester(type(actual))
        return tester.objects_are_equal(actual=actual, expected=expected, config=config)



r"""Define the reducer base class."""

from __future__ import annotations

__all__ = ["BaseBasicReducer", "BaseReducer", "EmptySequenceError"]

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

T = TypeVar("T", bound=Sequence[float])


class EmptySequenceError(Exception):
    r"""Raise when the sequence is empty because it is not possible to
    reduce an empty sequence."""


class BaseReducer(ABC, Generic[T]):
    r"""Define the base class to implement a reducer.

    Example:
        ```pycon
        >>> from coola.reducer import NumpyReducer
        >>> reducer = NumpyReducer()
        >>> reducer.max([-2, -1, 0, 1, 2])
        2
        >>> reducer.median([-2, -1, 0, 1, 2])
        0.0
        >>> reducer.sort([2, 1, -2, 3, 0])
        [-2, 0, 1, 2, 3]

        ```
    """

    @abstractmethod
    def max(self, values: T) -> int | float:
        r"""Compute the maximum value.

        Args:
            values: The values.

        Returns:
            The maximum value.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.max([-2, -1, 0, 1, 2])
            2

            ```
        """

    @abstractmethod
    def mean(self, values: T) -> float:
        r"""Compute the mean value.

        Args:
            values: The values.

        Returns:
            The mean value.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.mean([-2, -1, 0, 1, 2])
            0.0

            ```
        """

    @abstractmethod
    def median(self, values: T) -> int | float:
        r"""Compute the median value.

        Args:
            values: The values.

        Returns:
            The median value.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.median([-2, -1, 0, 1, 2])
            0

            ```
        """

    @abstractmethod
    def min(self, values: T) -> int | float:
        r"""Compute the minimum value.

        Args:
            values: The values.

        Returns:
            The minimum value.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.min([-2, -1, 0, 1, 2])
            -2

            ```
        """

    @abstractmethod
    def quantile(self, values: T, quantiles: Sequence[float]) -> list[float]:
        r"""Compute the quantiles.

        Args:
            values: The values.
            quantiles (sequence of float): The quantile
                values in the range ``[0, 1]``.

        Returns:
            The quantiles.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.quantile([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], (0.2, 0.5, 0.9))
            [2.0, 5.0, 9.0]

            ```
        """

    @abstractmethod
    def sort(self, values: T, descending: bool = False) -> list[int | float]:
        r"""Sorts the values.

        Args:
            values: The values.
            descending: The sorting order.

        Returns:
            The sorted values.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.sort([2, 1, -2, 3, 0])
            [-2, 0, 1, 2, 3]
            >>> reducer.sort([2, 1, -2, 3, 0], descending=True)
            [3, 2, 1, 0, -2]

            ```
        """

    @abstractmethod
    def std(self, values: T) -> float:
        r"""Compute the standard deviation.

        Args:
            values: The values.

        Returns:
            The standard deviation.

        Raises:
            EmptySequenceError: if the input sequence is empty.

        Example:
            ```pycon
            >>> from coola.reducer import TorchReducer
            >>> reducer = TorchReducer()
            >>> reducer.std([-2, -1, 0, 1, 2])
            1.581138...

            ```
        """


class BaseBasicReducer(BaseReducer[T]):
    r"""Extension of ``BaseReducer`` to check if the input sequence is
    empty before to call the reduction methods."""

    @abstractmethod
    def _is_empty(self, values: T) -> bool:
        r"""Return ``True`` if the values are empty.

        Returns:
            ``True`` if the values are empty, ``False`` otherwise.
        """

    def max(self, values: T) -> int | float:
        if self._is_empty(values):
            msg = "Cannot compute the maximum because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._max(values)

    @abstractmethod
    def _max(self, values: T) -> int | float:
        r"""Compute the maximum value.

        Args:
            values: The values.

        Returns:
            The maximum value.
        """

    def mean(self, values: T) -> int | float:
        if self._is_empty(values):
            msg = "Cannot compute the mean because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._mean(values)

    @abstractmethod
    def _mean(self, values: T) -> float:
        r"""Compute the mean value.

        Args:
            values: The values.

        Returns:
            The mean value.
        """

    def median(self, values: T) -> int | float:
        if self._is_empty(values):
            msg = "Cannot compute the median because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._median(values)

    @abstractmethod
    def _median(self, values: T) -> int | float:
        r"""Compute the median value.

        Args:
            values: The values.

        Returns:
            The median value.
        """

    def min(self, values: T) -> int | float:
        if self._is_empty(values):
            msg = "Cannot compute the minimum because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._min(values)

    @abstractmethod
    def _min(self, values: T) -> int | float:
        r"""Compute the minimum value.

        Args:
            values: The values.

        Returns:
            The minimum value.
        """

    def quantile(self, values: T, quantiles: Sequence[float]) -> list[float]:
        if self._is_empty(values):
            msg = "Cannot compute the quantiles because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._quantile(values, quantiles)

    @abstractmethod
    def _quantile(self, values: T, quantiles: Sequence[float]) -> list[float]:
        r"""Compute the quantiles.

        Args:
            values: The values.
            quantiles: The quantile values in the
                range ``[0, 1]``.

        Returns:
            The quantiles.
        """

    def std(self, values: T) -> float:
        if self._is_empty(values):
            msg = "Cannot compute the standard deviation because the sequence is empty"
            raise EmptySequenceError(msg)
        return self._std(values)

    @abstractmethod
    def _std(self, values: T) -> float:
        r"""Compute the standard deviation.

        Args:
            values: The values.

        Returns:
            The standard deviation.
        """



r"""Implement the default summarizer."""

from __future__ import annotations

__all__ = ["DefaultSummarizer"]

from typing import TYPE_CHECKING

from coola.summary.base import BaseSummarizer

if TYPE_CHECKING:
    from coola.summary.registry import SummarizerRegistry


class DefaultSummarizer(BaseSummarizer[object]):
    r"""Implement the default summarizer.

    Args:
        max_characters: The maximum number of characters to
            show. If a negative value is provided, all the characters
            are shown.

    Example:
        ```pycon
        >>> from coola.summary import SummarizerRegistry, DefaultSummarizer
        >>> registry = SummarizerRegistry()
        >>> summarizer = DefaultSummarizer()
        >>> print(summarizer.summarize(1, registry))
        <class 'int'> 1

        ```
    """

    def __init__(self, max_characters: int = -1) -> None:
        self._max_characters = max_characters

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(max_characters={self._max_characters:,})"

    def equal(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        return self._max_characters == other._max_characters

    def summarize(
        self,
        data: object,
        registry: SummarizerRegistry,  # noqa: ARG002
        depth: int = 0,
        max_depth: int = 1,
    ) -> str:
        if depth >= max_depth:
            return self._summarize(str(data))
        return f"{type(data)} {self._summarize(str(data))}"

    def _summarize(self, value: str) -> str:
        if self._max_characters >= 0 and len(value) > self._max_characters:
            value = value[: self._max_characters] + "..."
        return value



r"""Define the default transformer for mapping data (dict)."""

from __future__ import annotations

__all__ = ["MappingTransformer"]

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from coola.recursive.base import BaseTransformer

if TYPE_CHECKING:
    from collections.abc import Callable

    from coola.recursive.registry import TransformerRegistry


class MappingTransformer(BaseTransformer[Mapping[Any, Any]]):
    r"""Transformer for mapping types that recursively transforms values.

    This transformer handles dict-like mapping structures by recursively
    transforming all values while preserving keys unchanged. After
    transformation, it reconstructs the mapping using its original type
    (dict, OrderedDict, defaultdict, etc.), maintaining the mapping's
    specific characteristics and behavior.

    Notes:
        - Keys are never transformed, only values are processed recursively
        - The original mapping type is preserved in the output
        - Nested mappings and other containers in values are handled recursively
        - Empty mappings are preserved as empty mappings of the same type

    Example:
        ```pycon
        >>> from coola.recursive import MappingTransformer, TransformerRegistry, DefaultTransformer
        >>> registry = TransformerRegistry({object: DefaultTransformer()})
        >>> transformer = MappingTransformer()
        >>> transformer
        MappingTransformer()
        >>> # Transform simple dict values
        >>> transformer.transform({"a": 1, "b": 2}, func=str, registry=registry)
        {'a': '1', 'b': '2'}
        >>> # Keys remain unchanged
        >>> transformer.transform({1: "x", 2: "y"}, func=str.upper, registry=registry)
        {1: 'X', 2: 'Y'}
        >>> # Nested structures in values are handled recursively
        >>> transformer.transform({"nums": [1, 2, 3], "text": "hello"}, func=str, registry=registry)
        {'nums': '[1, 2, 3]', 'text': 'hello'}
        >>> # Empty mappings are preserved
        >>> transformer.transform({}, func=str, registry=registry)
        {}

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def transform(
        self,
        data: Mapping[Any, Any],
        func: Callable[[Any], Any],
        registry: TransformerRegistry,
    ) -> Mapping[Any, Any]:
        # Transform all values recursively using the registry
        transformed = {key: registry.transform(value, func) for key, value in data.items()}

        # Rebuild with original type to preserve mapping characteristics
        return type(data)(transformed)



r"""Define the BFS child finder class for handling iterable data."""

from __future__ import annotations

__all__ = ["MappingChildFinder"]

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from coola.iterator.bfs.base import BaseChildFinder

if TYPE_CHECKING:
    from collections.abc import Iterator


class MappingChildFinder(BaseChildFinder[Mapping[Any, Any]]):
    r"""Child finder for mapping objects.

    This child finder handles mapping objects by yielding the values
    from the mapping. It works with dictionaries and any
    object implementing the Mapping protocol.

    Example:
        ```pycon
        >>> from coola.iterator.bfs import MappingChildFinder
        >>> child_finder = MappingChildFinder()
        >>> list(child_finder.find_children({"a": 1, "b": 2}))
        [1, 2]
        >>> list(child_finder.find_children({"a": {"b": 1, "c": 2}, "d": 3}))
        [{'b': 1, 'c': 2}, 3]

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def find_children(self, data: Mapping[Any, Any]) -> Iterator[Any]:
        yield from data.values()



r"""Define the default DFS iterator class for handling leaf nodes in
traversal."""

from __future__ import annotations

__all__ = ["DefaultIterator"]

from typing import TYPE_CHECKING, Any

from coola.iterator.dfs.base import BaseIterator

if TYPE_CHECKING:
    from collections.abc import Iterator

    from coola.iterator.dfs.registry import IteratorRegistry


class DefaultIterator(BaseIterator[Any]):
    r"""Default iterator for depth-first search traversal of leaf nodes.

    This iterator serves as the fallback handler for objects that don't have
    a specialized iterator registered. It treats the input data as a leaf node
    and yields it directly without further traversal.

    The DefaultIterator is typically used for:
    - Primitive types (int, float, str, bool, None)
    - Objects without internal structure to traverse
    - Terminal nodes in a data structure

    Example:
        ```pycon
        >>> from coola.iterator.dfs import IteratorRegistry, DefaultIterator
        >>> iterator = DefaultIterator()
        >>> registry = IteratorRegistry()
        >>> list(iterator.iterate(42, registry))
        [42]
        >>> list(iterator.iterate("hello", registry))
        ['hello']

        ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def iterate(
        self,
        data: object,
        registry: IteratorRegistry,  # noqa: ARG002
    ) -> Iterator[Any]:
        yield data



r"""Define the manager registry for random number generator state
management.

This module provides a registry system that manages random number
generator managers, allowing coordinated control of multiple RNG states
across different libraries (e.g., Python's random module, PyTorch,
NumPy).
"""

from __future__ import annotations

__all__ = ["RandomManagerRegistry"]

from typing import TYPE_CHECKING, Any

from coola.random.base import BaseRandomManager
from coola.registry import Registry
from coola.utils.format import repr_indent, repr_mapping, str_indent, str_mapping

if TYPE_CHECKING:
    from collections.abc import Mapping


class RandomManagerRegistry(BaseRandomManager):
    """Registry that manages random number generator managers.

    This registry maintains a mapping from string keys to random manager instances,
    enabling centralized control of random number generator states across multiple
    libraries. It provides methods to seed all managers, get and set their states,
    and check for registered managers.

    Args:
        initial_state: Optional initial mapping of string keys to managers.
            If provided, the state is copied to prevent external mutations.

    Attributes:
        _state: Internal registry mapping keys to random managers

    Example:
        Basic usage with a random manager:

        ```pycon
        >>> from coola.random import RandomManagerRegistry, RandomRandomManager
        >>> registry = RandomManagerRegistry({"random": RandomRandomManager()})
        >>> registry
        RandomManagerRegistry(
          (state): Registry(
              (random): RandomRandomManager()
            )
        )
        >>> registry.manual_seed(42)

        ```
    """

    def __init__(self, initial_state: dict[str, BaseRandomManager] | None = None) -> None:
        self._state: Registry[str, BaseRandomManager] = Registry[str, BaseRandomManager](
            initial_state
        )

    def __repr__(self) -> str:
        state = repr_indent(repr_mapping({"state": self._state}))
        return f"{self.__class__.__qualname__}(\n  {state}\n)"

    def __str__(self) -> str:
        state = str_indent(str_mapping({"state": self._state}))
        return f"{self.__class__.__qualname__}(\n  {state}\n)"

    def register(
        self,
        key: str,
        manager: BaseRandomManager,
        exist_ok: bool = False,
    ) -> None:
        """Register a random manager with a given key.

        This method associates a manager instance with a specific string key.
        The manager will be used to control the random number generator state
        for its corresponding library.

        Args:
            key: The string key to register (e.g., "random", "torch", "numpy")
            manager: The random manager instance that handles RNG state
            exist_ok: If False (default), raises an error if the key is already
                registered. If True, overwrites the existing registration silently.

        Raises:
            RuntimeError: If the key is already registered and exist_ok is False

        Example:
            ```pycon
            >>> from coola.random import RandomManagerRegistry, RandomRandomManager
            >>> registry = RandomManagerRegistry()
            >>> registry.register("random", RandomRandomManager())
            >>> registry.has_manager("random")
            True

            ```
        """
        self._state.register(key, manager, exist_ok=exist_ok)

    def register_many(
        self,
        mapping: Mapping[str, BaseRandomManager],
        exist_ok: bool = False,
    ) -> None:
        """Register multiple random managers at once.

        This is a convenience method for bulk registration that internally calls
        register() for each key-manager pair.

        Args:
            mapping: Dictionary mapping string keys to random manager instances
            exist_ok: If False (default), raises an error if any key is already
                registered. If True, overwrites existing registrations silently.

        Raises:
            RuntimeError: If any key is already registered and exist_ok is False

        Example:
            ```pycon
            >>> from coola.random import (
            ...     RandomManagerRegistry,
            ...     RandomRandomManager,
            ...     TorchRandomManager,
            ... )
            >>> registry = RandomManagerRegistry()
            >>> registry.register_many(
            ...     {
            ...         "random": RandomRandomManager(),
            ...         "torch": TorchRandomManager(),
            ...     }
            ... )
            >>> registry
            RandomManagerRegistry(
              (state): Registry(
                  (random): RandomRandomManager()
                  (torch): TorchRandomManager()
                )
            )

            ```
        """
        self._state.register_many(mapping, exist_ok=exist_ok)

    def has_manager(self, key: str) -> bool:
        """Check if a random manager is registered for the given key.

        Args:
            key: The string key to check

        Returns:
            True if a manager is registered for this key, False otherwise

        Example:
            ```pycon
            >>> from coola.random import RandomManagerRegistry, RandomRandomManager
            >>> registry = RandomManagerRegistry()
            >>> registry.register("random", RandomRandomManager())
            >>> registry.has_manager("random")
            True
            >>> registry.has_manager("torch")
            False

            ```
        """
        return key in self._state

    def get_rng_state(self) -> dict[str, Any]:
        return {key: value.get_rng_state() for key, value in self._state.items()}

    def manual_seed(self, seed: int) -> None:
        for value in self._state.values():
            value.manual_seed(seed)

    def set_rng_state(self, state: dict[str, Any]) -> None:
        for key, value in state.items():
            self._state[key].set_rng_state(value)



r"""Define the abstract base class for depth-first search iterators."""

from __future__ import annotations

__all__ = ["BaseIterator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

    from coola.iterator.dfs.registry import IteratorRegistry

T = TypeVar("T")


class BaseIterator(ABC, Generic[T]):
    r"""Abstract base class for depth-first search iterators.

    This class defines the interface that all DFS iterators must implement.
    Iterators are responsible for traversing specific data types and yielding
    their elements during depth-first search traversal. Custom iterators can
    be registered with an ``IteratorRegistry`` to handle specific data types.

    The generic type parameter ``T`` indicates the type of data this iterator
    is designed to handle, though the ``iterate`` method accepts ``Any`` for
    flexibility.

    Notes:
        - Subclasses must implement the ``iterate`` method.
        - For container types, use ``registry.iterate()`` to recursively
          traverse nested structures.
        - For leaf types, simply yield the data directly.

    Example:
        ```pycon
        >>> from coola.iterator.dfs import IteratorRegistry, DefaultIterator
        >>> iterator = DefaultIterator()
        >>> registry = IteratorRegistry()
        >>> list(iterator.iterate(42, registry))
        [42]
        >>> list(iterator.iterate("hello", registry))
        ['hello']

        ```
    """

    @abstractmethod
    def iterate(self, data: T, registry: IteratorRegistry) -> Iterator[Any]:
        r"""Traverse the data structure and yield elements depth-first.

        This method defines how the iterator traverses its associated data type.
        For container or composite types, it should recursively traverse nested
        elements using ``registry.iterate()`` to delegate to appropriate
        iterators. For leaf types, it should yield the data directly.

        Args:
            data: The data structure to traverse. While typed as ``T`` for
                flexibility, implementations typically expect a specific type
                corresponding to the iterator's purpose.
            registry: The iterator registry used to resolve and dispatch
                iterators for nested data structures. Use ``registry.iterate()``
                to recursively traverse nested elements.

        Yields:
            Elements found during depth-first traversal. The exact type and
            nature of yielded elements depends on the specific iterator
            implementation and traversal strategy.

        Notes:
            - The registry parameter should be used to maintain consistent
              traversal behavior across different data types.
            - Implementations should handle the specific structure of their
              target data type appropriately.

        Example:
            ```pycon
            >>> from coola.iterator.dfs import IteratorRegistry, DefaultIterator
            >>> iterator = DefaultIterator()
            >>> registry = IteratorRegistry()
            >>> list(iterator.iterate(42, registry))
            [42]
            >>> list(iterator.iterate("hello", registry))
            ['hello']

            ```
        """



r"""Implement handlers for ``numpy.ndarray``s."""

from __future__ import annotations

__all__ = ["NumpyArrayEqualHandler"]

import logging
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_value_difference
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.utils.imports import is_numpy_available

if TYPE_CHECKING or is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    from coola.utils.fallback.numpy import numpy as np

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class NumpyArrayEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two NumPy arrays are equal.

    This handler returns ``True`` if the two arrays are equal,
    otherwise ``False``. This handler is designed to be used at
    the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import NumpyArrayEqualHandler
        >>> config = EqualityConfig()
        >>> handler = NumpyArrayEqualHandler()
        >>> handler.handle(np.ones((2, 3)), np.ones((2, 3)), config)
        True
        >>> handler.handle(np.ones((2, 3)), np.zeros((2, 3)), config)
        False

        ```
    """

    def handle(
        self,
        actual: np.ndarray,
        expected: np.ndarray,
        config: EqualityConfig,
    ) -> bool:
        object_equal = array_equal(actual, expected, config)
        if config.show_difference and not object_equal:
            logger.info(format_value_difference(actual, expected, name="numpy.ndarrays"))
        return object_equal


def array_equal(array1: np.ndarray, array2: np.ndarray, config: EqualityConfig) -> bool:
    r"""Indicate if the two arrays are equal within a tolerance.

    Args:
        array1: The first array to compare.
        array2: The second array to compare.
        config: The equality configuration.

    Returns:
        ``True`` if the two arrays are equal within a tolerance,
            otherwise ``False``.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler.numpy import array_equal
        >>> config = EqualityConfig()
        >>> array_equal(np.ones((2, 3)), np.ones((2, 3)), config)
        True
        >>> array_equal(np.ones((2, 3)), np.zeros((2, 3)), config)
        False

        ```
    """
    if (config.atol > 0 or config.rtol > 0) and is_numeric_array(array1):
        return np.allclose(
            array1, array2, rtol=config.rtol, atol=config.atol, equal_nan=config.equal_nan
        )
    return np.array_equal(array1, array2, equal_nan=config.equal_nan)


def is_numeric_array(array: np.ndarray) -> bool:
    r"""Indicate if the input array is a numeric array or not.

    Args:
        array: The input array.

    Returns:
        ``True`` if the input array is a numeric array, otherwise ``False``.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.handler.numpy import is_numeric_array
        >>> is_numeric_array(np.ones((2, 3)))
        True
        >>> is_numeric_array(np.array(["polar", "bear", "meow"]))
        False

        ```
    """
    return array.dtype.kind in {"?", "b", "B", "i", "u", "f", "c"}



r"""Contain some utility functions to manipulate mappings."""

from __future__ import annotations

__all__ = ["get_first_value", "remove_keys_starting_with", "to_flat_dict"]

from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from collections.abc import Mapping

T = TypeVar("T")


def get_first_value(data: Mapping[Any, T]) -> T:
    r"""Get the first value of a mapping.

    Args:
        data: The input mapping.

    Returns:
        The first value in the mapping.

    Raises:
        ValueError: if the mapping is empty.

    Example:
        ```pycon
        >>> from coola.nested import get_first_value
        >>> get_first_value({"key1": 1, "key2": 2})
        1

        ```
    """
    if not data:
        msg = "First value cannot be returned because the mapping is empty"
        raise ValueError(msg)
    return data[next(iter(data))]


def to_flat_dict(
    data: object,
    prefix: str | None = None,
    separator: str = ".",
    to_str: type[Any] | tuple[type[Any], ...] | None = None,
) -> dict[str, Any]:
    r"""Return a flat representation of a nested dict with the dot
    format.

    Args:
        data: The nested dict to flat.
        prefix: The prefix to use to generate the name of the key.
            ``None`` means no prefix.
        separator: The separator to concatenate keys of nested
            collections.
        to_str: The data types which will not be flattened out,
            instead converted to string.

    Returns:
        The flatted dictionary.

    Example:
        ```pycon
        >>> from coola.nested import to_flat_dict
        >>> data = {
        ...     "str": "def",
        ...     "module": {
        ...         "component": {
        ...             "float": 3.5,
        ...             "int": 2,
        ...         },
        ...     },
        ... }
        >>> to_flat_dict(data)
        {'str': 'def', 'module.component.float': 3.5, 'module.component.int': 2}
        >>> # Example with lists (also works with tuple)
        >>> data = {
        ...     "module": [[1, 2, 3], {"bool": True}],
        ...     "str": "abc",
        ... }
        >>> to_flat_dict(data)
        {'module.0.0': 1, 'module.0.1': 2, 'module.0.2': 3, 'module.1.bool': True, 'str': 'abc'}
        >>> # Example with lists with to_str=(list) (also works with tuple)
        >>> data = {
        ...     "module": [[1, 2, 3], {"bool": True}],
        ...     "str": "abc",
        ... }
        >>> to_flat_dict(data)
        {'module.0.0': 1, 'module.0.1': 2, 'module.0.2': 3, 'module.1.bool': True, 'str': 'abc'}

        ```
    """
    flat_dict = {}
    to_str = to_str or ()
    if isinstance(data, to_str):
        flat_dict[prefix] = str(data)
    elif isinstance(data, dict):
        for key, value in data.items():
            flat_dict.update(
                to_flat_dict(
                    value,
                    prefix=f"{prefix}{separator}{key}" if prefix else key,
                    separator=separator,
                    to_str=to_str,
                )
            )
    elif isinstance(data, (list, tuple)):
        for i, value in enumerate(data):
            flat_dict.update(
                to_flat_dict(
                    value,
                    prefix=f"{prefix}{separator}{i}" if prefix else str(i),
                    separator=separator,
                    to_str=to_str,
                )
            )
    else:
        flat_dict[prefix] = data
    return flat_dict


def remove_keys_starting_with(mapping: Mapping[Any, Any], prefix: str) -> dict[Any, Any]:
    r"""Remove the keys that start with a given prefix.

    Args:
        mapping: The original mapping.
        prefix: The prefix used to filter the keys.

    Returns:
        A new dict without the removed keys.

    Example:
        ```pycon
        >>> from coola.nested import remove_keys_starting_with
        >>> remove_keys_starting_with(
        ...     {"key": 1, "key.abc": 2, "abc": 3, "abc.key": 4, 1: 5, (2, 3): 6},
        ...     "key",
        ... )
        {'abc': 3, 'abc.key': 4, 1: 5, (2, 3): 6}

        ```
    """
    out = {}
    for key, value in mapping.items():
        if isinstance(key, str) and key.startswith(prefix):
            continue
        out[key] = value
    return out



r"""Implement some utility functions for ``torch.Tensor``s."""

from __future__ import annotations

__all__ = ["get_available_devices", "is_cuda_available", "is_mps_available", "to_tensor"]

from functools import lru_cache
from typing import TYPE_CHECKING

from coola.utils.imports import is_numpy_available, is_torch_available

if TYPE_CHECKING or is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    from coola.utils.fallback.numpy import numpy as np

if TYPE_CHECKING or is_torch_available():
    import torch
else:  # pragma: no cover
    from coola.utils.fallback.torch import torch

if TYPE_CHECKING:
    from collections.abc import Sequence


@lru_cache(1)
def get_available_devices() -> tuple[str, ...]:
    r"""Get the available PyTorch devices on the machine.

    Returns:
        The available devices.

    Example:
        ```pycon
        >>> from coola.utils.tensor import get_available_devices
        >>> get_available_devices()
        ('cpu'...)

        ```
    """
    devices = ["cpu"]
    if is_cuda_available():
        devices.append("cuda:0")
    if is_mps_available():
        devices.append("mps:0")
    return tuple(devices)


@lru_cache(1)
def is_cuda_available() -> bool:
    r"""Indicate if CUDA is currently available.

    Returns:
        A boolean indicating if CUDA is currently available.

    Example:
        ```pycon
        >>> from coola.utils.tensor import is_cuda_available
        >>> is_cuda_available()

        ```
    """
    return is_torch_available() and torch.cuda.is_available()


@lru_cache(1)
def is_mps_available() -> bool:
    r"""Indicate if MPS is currently available.

    Returns:
        A boolean indicating if MPS is currently available.

    Example:
        ```pycon
        >>> from coola.utils.tensor import is_mps_available
        >>> is_mps_available()

        ```
    """
    if not is_torch_available():
        return False
    try:
        torch.ones(1, device="mps")
    except RuntimeError:
        return False
    return True


def to_tensor(data: Sequence[int | float] | torch.Tensor | np.ndarray) -> torch.Tensor:
    r"""Convert the input to a ``torch.Tensor``.

    Args:
        data: The data to convert to a tensor.

    Returns:
        A tensor.

    Example:
        ```pycon
        >>> from coola.utils.tensor import to_tensor
        >>> x = to_tensor([1, 2, 3, 4, 5])
        >>> x
        tensor([1, 2, 3, 4, 5])
        >>> import numpy as np
        >>> x = to_tensor(np.array([1, 2, 3, 4, 5]))
        >>> x
        tensor([1, 2, 3, 4, 5])

        ```
    """
    if is_numpy_available() and isinstance(data, np.ndarray):
        return torch.from_numpy(data)
    if not torch.is_tensor(data):
        return torch.as_tensor(data)
    return data



r"""Contain the equality testers to check if two objects are equal or
not.

This module provides a comprehensive system for comparing objects of different types
using a registry-based approach with specialized equality testers. The testers use
the chain of responsibility pattern with handlers to perform various equality checks.

Key components:
    - BaseEqualityTester: Abstract base class for all equality testers
    - EqualityTesterRegistry: Registry that dispatches to appropriate testers by type
    - Specialized testers for Python built-ins (list, dict, int, float, etc.)
    - Testers for third-party libraries (NumPy, PyTorch, Pandas, Polars, JAX, xarray, PyArrow)

The default registry is pre-configured with testers for common types and can be
extended with custom testers using register_equality_testers().

Example:
    ```pycon
    >>> from coola.equality.tester import get_default_registry
    >>> from coola.equality.config import EqualityConfig
    >>> registry = get_default_registry()
    >>> config = EqualityConfig()
    >>> # Compare lists with nested structures
    >>> registry.objects_are_equal([1, {"a": 2}], [1, {"a": 2}], config)
    True

    ```
"""

from __future__ import annotations

__all__ = [
    "BaseEqualityTester",
    "DefaultEqualityTester",
    "EqualEqualityTester",
    "EqualNanEqualityTester",
    "EqualityTesterRegistry",
    "HandlerEqualityTester",
    "JaxArrayEqualityTester",
    "MappingEqualityTester",
    "NumpyArrayEqualityTester",
    "NumpyMaskedArrayEqualityTester",
    "PandasDataFrameEqualityTester",
    "PandasSeriesEqualityTester",
    "PolarsDataFrameEqualityTester",
    "PolarsLazyFrameEqualityTester",
    "PolarsSeriesEqualityTester",
    "PyarrowEqualityTester",
    "ScalarEqualityTester",
    "SequenceEqualityTester",
    "TorchPackedSequenceEqualityTester",
    "TorchTensorEqualityTester",
    "XarrayDataArrayEqualityTester",
    "XarrayDatasetEqualityTester",
    "XarrayVariableEqualityTester",
    "get_default_registry",
    "register_equality_testers",
]

from coola.equality.tester.base import BaseEqualityTester
from coola.equality.tester.collection import (
    MappingEqualityTester,
    SequenceEqualityTester,
)
from coola.equality.tester.default import DefaultEqualityTester
from coola.equality.tester.equal import EqualEqualityTester, EqualNanEqualityTester
from coola.equality.tester.handler import HandlerEqualityTester
from coola.equality.tester.interface import (
    get_default_registry,
    register_equality_testers,
)
from coola.equality.tester.jax import JaxArrayEqualityTester
from coola.equality.tester.numpy import (
    NumpyArrayEqualityTester,
    NumpyMaskedArrayEqualityTester,
)
from coola.equality.tester.pandas import (
    PandasDataFrameEqualityTester,
    PandasSeriesEqualityTester,
)
from coola.equality.tester.polars import (
    PolarsDataFrameEqualityTester,
    PolarsLazyFrameEqualityTester,
    PolarsSeriesEqualityTester,
)
from coola.equality.tester.pyarrow import PyarrowEqualityTester
from coola.equality.tester.registry import EqualityTesterRegistry
from coola.equality.tester.scalar import ScalarEqualityTester
from coola.equality.tester.torch import (
    TorchPackedSequenceEqualityTester,
    TorchTensorEqualityTester,
)
from coola.equality.tester.xarray import (
    XarrayDataArrayEqualityTester,
    XarrayDatasetEqualityTester,
    XarrayVariableEqualityTester,
)



r"""Define the public interface to recursively apply a function to all
items in nested data."""

from __future__ import annotations

__all__ = ["objects_are_allclose", "objects_are_equal"]

from typing import TYPE_CHECKING

from coola.equality.config import EqualityConfig
from coola.equality.tester.interface import get_default_registry

if TYPE_CHECKING:
    from coola.equality.tester.registry import EqualityTesterRegistry


def objects_are_allclose(
    actual: object,
    expected: object,
    *,
    rtol: float = 1e-5,
    atol: float = 1e-8,
    equal_nan: bool = False,
    show_difference: bool = False,
    max_depth: int = 1000,
    registry: EqualityTesterRegistry | None = None,
) -> bool:
    r"""Indicate if two objects are equal within a tolerance.

    Args:
        actual: The actual input.
        expected: The expected input.
        rtol: The relative tolerance parameter. Must be non-negative.
        atol: The absolute tolerance parameter. Must be non-negative.
        equal_nan: If ``True``, then two ``NaN``s  will be considered
            as equal.
        show_difference: If ``True``, it shows a difference between
            the two objects if they are different. This parameter is
            useful to find the difference between two objects.
        max_depth: Maximum recursion depth for nested comparisons.
            Must be positive. Defaults to 1000.
        registry: The registry with the equality tester to use.

    Returns:
        ``True`` if the two objects are (element-wise) equal within a
            tolerance, otherwise ``False``

    Raises:
        ValueError: if ``rtol`` or ``atol`` is negative.
        RecursionError: if recursion depth exceeds ``max_depth``.

    Example:
        ```pycon

        >>> import torch
        >>> from coola.equality import objects_are_allclose
        >>> objects_are_allclose(
        ...     [torch.ones(2, 3), torch.zeros(2)],
        ...     [torch.ones(2, 3), torch.zeros(2)],
        ... )
        True
        >>> objects_are_allclose(
        ...     [torch.ones(2, 3), torch.ones(2)],
        ...     [torch.ones(2, 3), torch.zeros(2)],
        ... )
        False
        >>> objects_are_allclose(
        ...     [torch.ones(2, 3) + 1e-7, torch.ones(2)],
        ...     [torch.ones(2, 3), torch.ones(2) - 1e-7],
        ...     rtol=0,
        ...     atol=1e-8,
        ... )
        False

        ```
    """
    if rtol < 0:
        msg = f"rtol must be non-negative, but got {rtol}"
        raise ValueError(msg)
    if atol < 0:
        msg = f"atol must be non-negative, but got {atol}"
        raise ValueError(msg)
    if registry is None:
        registry = get_default_registry()
    config = EqualityConfig(
        registry=registry,
        show_difference=show_difference,
        equal_nan=equal_nan,
        atol=atol,
        rtol=rtol,
        max_depth=max_depth,
    )
    return registry.objects_are_equal(actual, expected, config)


def objects_are_equal(
    actual: object,
    expected: object,
    *,
    equal_nan: bool = False,
    show_difference: bool = False,
    max_depth: int = 1000,
    registry: EqualityTesterRegistry | None = None,
) -> bool:
    r"""Indicate if two objects are equal or not.

    Args:
        actual: The actual input.
        expected: The expected input.
        equal_nan: If ``True``, then two ``NaN``s  will be considered
            as equal.
        show_difference: If ``True``, it shows a difference between
            the two objects if they are different. This parameter is
            useful to find the difference between two objects.
        max_depth: Maximum recursion depth for nested comparisons.
            Must be positive. Defaults to 1000.
        registry: The registry with the equality tester to use.

    Returns:
        ``True`` if the two nested data are equal, otherwise
            ``False``.

    Raises:
        RecursionError: if recursion depth exceeds ``max_depth``.

    Example:
        ```pycon

        >>> import torch
        >>> from coola.equality import objects_are_equal
        >>> objects_are_equal(
        ...     [torch.ones(2, 3), torch.zeros(2)],
        ...     [torch.ones(2, 3), torch.zeros(2)],
        ... )
        True
        >>> objects_are_equal([torch.ones(2, 3), torch.ones(2)], [torch.ones(2, 3), torch.zeros(2)])
        False

        ```
    """
    if registry is None:
        registry = get_default_registry()
    config = EqualityConfig(
        registry=registry,
        show_difference=show_difference,
        equal_nan=equal_nan,
        max_depth=max_depth,
    )
    return registry.objects_are_equal(actual, expected, config)



r"""Implement utility functions to securely get passwords from users."""

from __future__ import annotations

__all__ = ["get_password"]

import getpass
import hmac
import sys
from typing import Final

MAX_PASSWORD_LENGTH: Final[int] = 1024


def get_password(username: str | None = None, *, confirm: bool = False) -> str:
    r"""Securely prompt for and retrieve a password from the user.

    This function uses getpass to securely prompt for password input without
    echoing characters to the terminal. The password undergoes validation
    for length constraints before being returned.

    Security features:
        - No password echoing to terminal
        - Length validation (non-empty, max bound)
        - Optional confirmation to prevent typos
        - Prevents use in non-interactive contexts
        - No logging or storage of password values

    Args:
        username: Optional username for prompt clarity.
        confirm: Require password confirmation if True.

    Returns:
        The password entered by the user (whitespace preserved).

    Raises:
        RuntimeError: If not running in an interactive terminal.
        KeyboardInterrupt, EOFError: If input is interrupted.
        ValueError: If password validation fails.
    """
    if not (sys.stdin.isatty() and sys.stderr.isatty()):
        msg = "Password input requires an interactive terminal"
        raise RuntimeError(msg)

    prompt = f"Enter password for user {username}: " if username else "Enter password: "

    password = getpass.getpass(prompt)

    if not password:
        msg = "Password cannot be empty"
        raise ValueError(msg)

    if len(password) > MAX_PASSWORD_LENGTH:
        msg = f"Password exceeds maximum length of {MAX_PASSWORD_LENGTH} characters"
        raise ValueError(msg)

    if confirm:
        confirmation = getpass.getpass("Confirm password: ")
        if not hmac.compare_digest(password, confirmation):
            msg = "Passwords do not match"
            raise ValueError(msg)

    return password



r"""Implement a handler to check if two objects have the same shape."""

from __future__ import annotations

__all__ = ["SameShapeHandler", "SupportsShape"]

import logging
from typing import TYPE_CHECKING, Protocol

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import format_shape_difference
from coola.equality.handler.mixin import HandlerEqualityMixin

if TYPE_CHECKING:
    from coola.equality.config import EqualityConfig


logger: logging.Logger = logging.getLogger(__name__)


class SupportsShape(Protocol):
    r"""Implement a protocol to represent objects with a ``shape``
    attribute.

    This protocol can be used to represent several objects like
    ``jax.numpy.ndarray``s, ``numpy.ndarray``s, ``pandas.DataFrame``,
    ``polars.DataFrame`` and ``torch.Tensor``s.
    """

    @property
    def shape(self) -> tuple[int, ...]:
        return ()  # pragma: no cover


class SameShapeHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same shape.

    This handler returns ``False`` if the two objects have different
    shapes, otherwise it passes the inputs to the next handler.
    The objects must have a ``shape`` attribute (e.g. ``object.shape``)
    which returns the shape of the object. This handler works on
    ``jax.numpy.ndarray``s, ``numpy.ndarray``s, ``pandas.DataFrame``,
    ``polars.DataFrame`` and ``torch.Tensor``s objects.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameShapeHandler, TrueHandler
        >>> config = EqualityConfig()
        >>> handler = SameShapeHandler(next_handler=TrueHandler())
        >>> handler.handle(np.ones((2, 3)), np.ones((2, 3)), config)
        True
        >>> handler.handle(np.ones((2, 3)), np.ones((3, 2)), config)
        False

        ```
    """

    def handle(
        self, actual: SupportsShape, expected: SupportsShape, config: EqualityConfig
    ) -> bool:
        if actual.shape != expected.shape:
            if config.show_difference:
                logger.info(format_shape_difference(actual.shape, expected.shape))
            return False
        return self._handle_next(actual, expected, config=config)



r"""Contain functionalities to compute a text summary of nested data
based on the type of data."""

from __future__ import annotations

__all__ = [
    "BaseCollectionSummarizer",
    "BaseSummarizer",
    "DefaultSummarizer",
    "MappingSummarizer",
    "NDArraySummarizer",
    "SequenceSummarizer",
    "SetSummarizer",
    "SummarizerRegistry",
    "TensorSummarizer",
    "get_default_registry",
    "register_summarizers",
    "summarize",
]

from coola.summary.base import BaseSummarizer
from coola.summary.collection import BaseCollectionSummarizer
from coola.summary.default import DefaultSummarizer
from coola.summary.interface import (
    get_default_registry,
    register_summarizers,
    summarize,
)
from coola.summary.mapping import MappingSummarizer
from coola.summary.numpy import NDArraySummarizer
from coola.summary.registry import SummarizerRegistry
from coola.summary.sequence import SequenceSummarizer
from coola.summary.set import SetSummarizer
from coola.summary.torch import TensorSummarizer



r"""Implement handlers for native Python objects."""

from __future__ import annotations

__all__ = [
    "FalseHandler",
    "ObjectEqualHandler",
    "SameAttributeHandler",
    "SameLengthHandler",
    "SameObjectHandler",
    "SameTypeHandler",
    "TrueHandler",
]

import logging
from typing import TYPE_CHECKING

from coola.equality.handler.base import BaseEqualityHandler
from coola.equality.handler.format import (
    format_type_difference,
    format_value_difference,
)
from coola.equality.handler.mixin import HandlerEqualityMixin
from coola.equality.handler.utils import handlers_are_equal

if TYPE_CHECKING:
    from collections.abc import Sized

    from coola.equality.config import EqualityConfig

logger: logging.Logger = logging.getLogger(__name__)


class FalseHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Implement a handler that always returns ``False``.

    This handler is designed to be used at the end of the chain of
    responsibility. This handler does not call the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import FalseHandler
        >>> config = EqualityConfig()
        >>> handler = FalseHandler()
        >>> handler.handle("abc", "abc", config)
        False
        >>> handler.handle("abc", "ABC", config)
        False

        ```
    """

    def handle(
        self,
        actual: object,  # noqa: ARG002
        expected: object,  # noqa: ARG002
        config: EqualityConfig,  # noqa: ARG002
    ) -> bool:
        return False


class TrueHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Implement a handler that always returns ``True``.

    This handler is designed to be used at the end of the chain of
    responsibility. This handler does not call the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import TrueHandler
        >>> config = EqualityConfig()
        >>> handler = TrueHandler()
        >>> handler.handle("abc", "abc", config)
        True
        >>> handler.handle("abc", "ABC", config)
        True

        ```
    """

    def handle(
        self,
        actual: object,  # noqa: ARG002
        expected: object,  # noqa: ARG002
        config: EqualityConfig,  # noqa: ARG002
    ) -> bool:
        return True


class ObjectEqualHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects are equal using the default equality
    operator ``==``.

    This handler returns ``True`` if the two objects are equal,
    otherwise ``False``. This handler is designed to be used at
    the end of the chain of responsibility. This handler does
    not call the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import ObjectEqualHandler
        >>> config = EqualityConfig()
        >>> handler = ObjectEqualHandler()
        >>> handler.handle(1, 1, config)
        True
        >>> handler.handle(1, "abc", config)
        False

        ```
    """

    def handle(
        self,
        actual: object,
        expected: object,
        config: EqualityConfig,
    ) -> bool:
        object_equal = actual == expected
        if config.show_difference and not object_equal:
            logger.info(format_value_difference(actual=actual, expected=expected))
        return object_equal


class SameAttributeHandler(BaseEqualityHandler):
    r"""Check if the two objects have the same attribute.

    This handler returns ``False`` if the two objects have different
    attributes, otherwise it passes the inputs to the next handler.
    The objects must have the attribute.

    Example:
        ```pycon
        >>> import numpy as np
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameAttributeHandler, TrueHandler
        >>> config = EqualityConfig()
        >>> handler = SameAttributeHandler(name="shape", next_handler=TrueHandler())
        >>> handler.handle(np.ones((2, 3)), np.ones((2, 3)), config)
        True
        >>> handler.handle(np.ones((2, 3)), np.ones((3, 2)), config)
        False

        ```
    """

    def __init__(self, name: str, next_handler: BaseEqualityHandler | None = None) -> None:
        super().__init__(next_handler=next_handler)
        self._name = name

    def __repr__(self) -> str:
        args = ""
        if self.next_handler:
            args += f", next_handler={self.next_handler}"
        return f"{self.__class__.__qualname__}(name={self._name}{args})"

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(name={self._name})"

    @property
    def name(self) -> str:
        return self._name

    def equal(self, other: object) -> bool:
        if type(other) is not type(self):
            return False
        if self.name != other.name:
            return False
        return handlers_are_equal(self.next_handler, other.next_handler)

    def handle(self, actual: object, expected: object, config: EqualityConfig) -> bool:
        value1 = getattr(actual, self._name)
        value2 = getattr(expected, self._name)
        if not config.registry.objects_are_equal(value1, value2, config):
            if config.show_difference:
                logger.info(f"objects have different {self._name}: {value1} vs {value2}")
            return False
        return self._handle_next(actual, expected, config=config)


class SameLengthHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same length.

    This handler returns ``False`` if the two objects have different
    lengths, otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameLengthHandler
        >>> config = EqualityConfig()
        >>> handler = SameLengthHandler()
        >>> handler.handle([1, 2, 3], [1, 2, 3, 4], config)
        False

        ```
    """

    def handle(
        self,
        actual: Sized,
        expected: Sized,
        config: EqualityConfig,
    ) -> bool:
        if len(actual) != len(expected):
            if config.show_difference:
                logger.info(f"objects have different lengths: {len(actual):,} vs {len(expected):,}")
            return False
        return self._handle_next(actual, expected, config=config)


class SameObjectHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects refer to the same object.

    This handler returns ``True`` if the two objects refer to the
    same object, otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameObjectHandler
        >>> config = EqualityConfig()
        >>> handler = SameObjectHandler()
        >>> handler.handle("abc", "abc", config)
        True

        ```
    """

    def handle(
        self,
        actual: object,
        expected: object,
        config: EqualityConfig,
    ) -> bool:
        if actual is expected:
            return True
        return self._handle_next(actual, expected, config=config)


class SameTypeHandler(HandlerEqualityMixin, BaseEqualityHandler):
    r"""Check if the two objects have the same type.

    This handler returns ``False`` if the two objects have different
    types, otherwise it passes the inputs to the next handler.

    Example:
        ```pycon
        >>> from coola.equality.config import EqualityConfig
        >>> from coola.equality.handler import SameTypeHandler
        >>> config = EqualityConfig()
        >>> handler = SameTypeHandler()
        >>> handler.handle(1, "abc", config)
        False

        ```
    """

    def handle(
        self,
        actual: object,
        expected: object,
        config: EqualityConfig,
    ) -> bool:
        if type(actual) is not type(expected):
            if config.show_difference:
                message = format_type_difference(type(actual), type(expected))
                logger.info(message)
            return False
        return self._handle_next(actual, expected, config=config)
